from typing import Union, Optional

import numpy as np
import pyvista as pv
from matplotlib import colors as mcolors
from pyvista import PolyData

from gempy_engine.core.data.raw_arrays_solution import RawArraysSolution
from gempy_viewer.core.scalar_data_type import ScalarDataType
from gempy.core.data.grid_modules import RegularGrid, Topography
from gempy_viewer.modules.plot_3d.vista import GemPyToVista


def plot_structured_grid(
        gempy_vista: GemPyToVista,
        regular_grid: RegularGrid,
        scalar_data_type: ScalarDataType,
        solution: RawArraysSolution,
        cmap: Union[mcolors.Colormap or str],
        active_scalar_field: Optional[str] = None,
        opacity=.25,
        **kwargs
):
    structured_grid: pv.StructuredGrid | pv.PolyData = create_regular_mesh(gempy_vista, regular_grid)

    # Set the scalar field-Activate it-getting cmap?
    structured_grid = set_scalar_data(
        structured_grid=structured_grid,
        data=solution,
        resolution = regular_grid.resolution,
        scalar_data_type=scalar_data_type
    )

    structured_grid = set_active_scalar_fields(
        structured_grid=structured_grid,
        active_scalar_field=active_scalar_field
    )
    topography_polydata: PolyData = gempy_vista.surface_poly.get('topography', None)
    if topography_polydata is not None:
        structured_grid = structured_grid.clip_surface(
            surface=topography_polydata,
            value=-10,
            crinkle=False,
            invert=True
        )

    add_regular_grid_mesh(
        gempy_vista=gempy_vista,
        structured_grid=structured_grid,
        cmap=cmap,
        opacity=opacity,  # BUG pass this as an argument
        **kwargs
    )


def add_regular_grid_mesh(
        gempy_vista: GemPyToVista,
        structured_grid: pv.StructuredGrid,
        cmap: Union[mcolors.Colormap or str],
        opacity: float,
        **kwargs
):
    gempy_vista.regular_grid_actor = gempy_vista.p.add_mesh(
        mesh=structured_grid,
        cmap=cmap,
        # ? scalars=main_scalar, if we prepare the structured grid do we need this arg?
        show_scalar_bar=True,
        scalar_bar_args=gempy_vista.scalar_bar_arguments,
        interpolate_before_map=True,
        opacity=opacity,
        clim=(0, cmap.N - 1),
        **kwargs
    )


def create_regular_mesh(gempy_vista: GemPyToVista, regular_grid: RegularGrid) -> pv.StructuredGrid:
    gempy_vista._grid_values = regular_grid.values

    if False:
        grid_3d = regular_grid.values.reshape(*regular_grid.resolution, 3).T
        regular_grid_mesh = pv.StructuredGrid(*grid_3d)
    else:
        x = np.linspace(regular_grid.extent[0], regular_grid.extent[1], regular_grid.resolution[0] + 1)
        y = np.linspace(regular_grid.extent[2], regular_grid.extent[3], regular_grid.resolution[1] + 1)
        z = np.linspace(regular_grid.extent[4], regular_grid.extent[5], regular_grid.resolution[2] + 1)

        x, y, z = np.meshgrid(x, y, z, indexing='ij')
        regular_grid_mesh = pv.StructuredGrid( x , y, z ) 
    return regular_grid_mesh


def _mask_topography(structured_grid: pv.StructuredGrid, topography: Topography) -> pv.StructuredGrid:
    # ? Obsolete? I am using pyvista clipping and seems to do the job very good.
    threshold = -100
    structured_grid.active_scalars[topography.topography_mask.ravel(order='C')] = threshold - 1

    # ? Is this messing up the data type?
    structured_grid: pv.StructuredGrid = structured_grid.threshold(
        value=threshold,
        method="upper"
    )

    return structured_grid


def set_scalar_data(
        data: RawArraysSolution,
        structured_grid: pv.StructuredGrid,
        resolution: np.ndarray,
        scalar_data_type: ScalarDataType,
) -> pv.StructuredGrid:
    
    def _convert_sol_array_to_fortran_order(array: np.ndarray) -> np.ndarray:
        return array.reshape(*resolution, order='C').ravel(order='F')
    # Substitute the madness of the previous if with match
    match scalar_data_type:
        case ScalarDataType.LITHOLOGY | ScalarDataType.ALL:
            structured_grid.cell_data['id'] = _convert_sol_array_to_fortran_order(data.lith_block - 1)
        case ScalarDataType.SCALAR_FIELD | ScalarDataType.ALL:
            scalar_field_ = 'sf_'
            for e in range(data.scalar_field_matrix.shape[0]):
                # TODO: Ideally we will have the group name instead the enumeration
                structured_grid[scalar_field_ + str(e)] = _convert_sol_array_to_fortran_order(data.scalar_field_matrix[e])
        case ScalarDataType.VALUES | ScalarDataType.ALL:
            scalar_field_ = 'values_'
            for e in range(data.values_matrix.shape[0]):
                structured_grid[scalar_field_ + str(e)] = _convert_sol_array_to_fortran_order(data.values_matrix[e])
        case _:
            raise ValueError(f'Unknown scalar data type: {scalar_data_type}')

    return structured_grid  # , cmap


def set_active_scalar_fields(structured_grid: pv.StructuredGrid, active_scalar_field: Optional[str]) -> pv.StructuredGrid:
    if active_scalar_field is None:
        active_scalar_field = structured_grid.array_names[0]

    if active_scalar_field == 'lith':
        active_scalar_field = 'id'

    # Set the scalar field active
    try:
        structured_grid.set_active_scalars(active_scalar_field)
    except ValueError:
        raise AttributeError('The scalar field provided does not exist. Please pass '
                             'a valid field: {}'.format(structured_grid.array_names))
    return structured_grid
   


"""
Modify after arviz 0.4!

Utilities for plotting."""
from itertools import product

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import xarray as xr


def make_2d(ary):
    """Convert any array into a 2d numpy array.

    In case the array is already more than 2 dimensional, will ravel the
    dimensions after the first.
    """
    dim_0, *_ = np.atleast_1d(ary).shape
    return ary.reshape(dim_0, -1, order="F")


def _scale_fig_size(figsize, textsize, rows=1, cols=1):
    """Scale figure properties according to rows and cols.

    Parameters
    ----------
    figsize : float or None
        Size of figure in inches
    textsize : float or None
        fontsize
    rows : int
        Number of rows
    cols : int
        Number of columns

    Returns
    -------
    figsize : float or None
        Size of figure in inches
    ax_labelsize : int
        fontsize for axes label
    titlesize : int
        fontsize for title
    xt_labelsize : int
        fontsize for axes ticks
    linewidth : int
        linewidth
    markersize : int
        markersize
    """
    params = mpl.rcParams
    rc_width, rc_height = tuple(params["figure.figsize"])
    rc_ax_labelsize = params["axes.labelsize"]
    rc_titlesize = params["axes.titlesize"]
    rc_xt_labelsize = params["xtick.labelsize"]
    rc_linewidth = params["lines.linewidth"]
    rc_markersize = params["lines.markersize"]
    if isinstance(rc_ax_labelsize, str):
        rc_ax_labelsize = 15
    if isinstance(rc_titlesize, str):
        rc_titlesize = 16
    if isinstance(rc_xt_labelsize, str):
        rc_xt_labelsize = 14

    if figsize is None:
        width, height = rc_width, rc_height
        sff = 1 if (rows == cols == 1) else 1.15
        width = width * cols * sff
        height = height * rows * sff
    else:
        width, height = figsize

    if textsize is not None:
        scale_factor = textsize / rc_xt_labelsize
    elif rows == cols == 1:
        scale_factor = ((width * height) / (rc_width * rc_height)) ** 0.5
    else:
        scale_factor = 1

    ax_labelsize = rc_ax_labelsize * scale_factor
    titlesize = rc_titlesize * scale_factor
    xt_labelsize = rc_xt_labelsize * scale_factor
    linewidth = rc_linewidth * scale_factor
    markersize = rc_markersize * scale_factor

    return (width, height), ax_labelsize, titlesize, xt_labelsize, linewidth, markersize


def get_bins(values):
    """
    Automatically compute the number of bins for discrete variables.

    Parameters
    ----------
    values = numpy array
        values

    Returns
    -------
    array with the bins

    Notes
    -----
    Computes the width of the bins by taking the maximun of the Sturges and the Freedman-Diaconis
    estimators. Acording to numpy `np.histogram` this provides good all around performance.

    The Sturges is a very simplistic estimator based on the assumption of normality of the data.
    This estimator has poor performance for non-normal data, which becomes especially obvious for
    large data sets. The estimate depends only on size of the data.

    The Freedman-Diaconis rule uses interquartile range (IQR) to estimate the binwidth.
    It is considered a robusts version of the Scott rule as the IQR is less affected by outliers
    than the standard deviation. However, the IQR depends on fewer points than the standard
    deviation, so it is less accurate, especially for long tailed distributions.
    """
    x_min = values.min().astype(int)
    x_max = values.max().astype(int)

    # Sturges histogram bin estimator
    bins_sturges = (x_max - x_min) / (np.log2(values.size) + 1)

    # The Freedman-Diaconis histogram bin estimator.
    iqr = np.subtract(*np.percentile(values, [75, 25]))  # pylint: disable=assignment-from-no-return
    bins_fd = 2 * iqr * values.size ** (-1 / 3)

    width = round(np.max([1, bins_sturges, bins_fd])).astype(int)

    return np.arange(x_min, x_max + width + 1, width)


def default_grid(n_items, max_cols=4, min_cols=3):  # noqa: D202
    """Make a grid for subplots.

    Tries to get as close to sqrt(n_items) x sqrt(n_items) as it can,
    but allows for custom logic

    Parameters
    ----------
    n_items : int
        Number of panels required
    max_cols : int
        Maximum number of columns, inclusive
    min_cols : int
        Minimum number of columns, inclusive

    Returns
    -------
    (int, int)
        Rows and columns, so that rows * columns >= n_items
    """

    def in_bounds(val):
        return np.clip(val, min_cols, max_cols)

    if n_items <= max_cols:
        return 1, n_items
    ideal = in_bounds(round(n_items ** 0.5))

    for offset in (0, 1, -1, 2, -2):
        cols = in_bounds(ideal + offset)
        rows, extra = divmod(n_items, cols)
        if extra == 0:
            return rows, cols
    return n_items // ideal + 1, ideal


def _create_axes_grid(length_plotters, rows, cols, **kwargs):
    """Create figure and axes for grids with multiple plots.

    Parameters
    ----------
    n_items : int
        Number of panels required
    rows : int
        Number of rows
    cols : int
        Number of columns

    Returns
    -------
    fig : matplotlib figure
    ax : matplotlib axes
    """
    kwargs.setdefault("constrained_layout", True)
    fig, ax = plt.subplots(rows, cols, **kwargs)
    ax = np.ravel(ax)
    extra = (rows * cols) - length_plotters
    if extra:
        for i in range(1, extra + 1):
            ax[-i].set_axis_off()
        ax = ax[:-extra]
    return fig, ax


def selection_to_string(selection):
    """Convert dictionary of coordinates to a string for labels.

    Parameters
    ----------
    selection : dict[Any] -> Any

    Returns
    -------
    str
        key1: value1, key2: value2, ...
    """
    return ", ".join(["{}".format(v) for _, v in selection.items()])


def make_label(var_name, selection, position="below"):
    """Consistent labelling for plots.

    Parameters
    ----------
    var_name : str
       Name of the variable

    selection : dict[Any] -> Any
        Coordinates of the variable
    position : whether to position the coordinates' label "below" (default) or "beside" the name
               of the variable

    Returns
    -------
    label
        A text representation of the label
    """
    if selection:
        sel = selection_to_string(selection)
        if position == "below":
            sep = "\n"
        elif position == "beside":
            sep = " "
    else:
        sep = sel = ""
    return "{}{}{}".format(var_name, sep, sel)


def purge_duplicates(list_in):
    """Remove duplicates from list while preserving order.

    Parameters
    ----------
    list_in: Iterable

    Returns
    -------
    list
        List of first occurences in order
    """
    _list = []
    for item in list_in:
        if item not in _list:
            _list.append(item)
    return _list


def xarray_var_iter(data, var_names=None, combined=False, skip_dims=None, reverse_selections=False):
    """Convert xarray data to an iterator over vectors.

    Iterates over each var_name and all of its coordinates, returning the 1d
    data.

    Parameters
    ----------
    data : xarray.Dataset
        Posterior data in an xarray

    var_names : iterator of strings (optional)
        Should be a subset of data.data_vars. Defaults to all of them.

    combined : bool
        Whether to combine chains or leave them separate

    skip_dims : set
        dimensions to not iterate over

    reverse_selections : bool
        Whether to reverse selections before iterating.

    Returns
    -------
    Iterator of (str, dict(str, any), np.array)
        The string is the variable name, the dictionary are coordinate names to values,
        and the array are the values of the variable at those coordinates.
    """
    if skip_dims is None:
        skip_dims = set()

    if combined:
        skip_dims = skip_dims.union({"chain", "draw"})
    else:
        skip_dims.add("draw")

    if var_names is None:
        if isinstance(data, xr.Dataset):
            var_names = list(data.data_vars)
        elif isinstance(data, xr.DataArray):
            var_names = [data.name]
            data = {data.name: data}

    for var_name in var_names:
        if var_name in data:
            new_dims = [dim for dim in data[var_name].dims if dim not in skip_dims]
            vals = [purge_duplicates(data[var_name][dim].values) for dim in new_dims]
            dims = [{k: v for k, v in zip(new_dims, prod)} for prod in product(*vals)]
            if reverse_selections:
                dims = reversed(dims)

            for selection in dims:
                yield var_name, selection, data[var_name].sel(**selection).values


def xarray_to_ndarray(data, *, var_names=None, combined=True):
    """Take xarray data and unpacks into variables and data into list and numpy array respectively.

    Assumes that chain and draw are in coordinates

    Parameters
    ----------
    data: xarray.DataSet
        Data in an xarray from an InferenceData object. Examples include posterior or sample_stats

    var_names: iter
        Should be a subset of data.data_vars not including chain and draws. Defaults to all of them

    combined: bool
        Whether to combine chain into one array

    Returns
    -------
    var_names: list
        List of variable names
    data: np.array
        Data values
    """
    unpacked_data, unpacked_var_names, = [], []

    # Merge chains and variables
    for var_name, selection, data_array in xarray_var_iter(
        data, var_names=var_names, combined=combined
    ):
        unpacked_data.append(data_array.flatten())
        unpacked_var_names.append(make_label(var_name, selection))

    return unpacked_var_names, np.array(unpacked_data)


def get_coords(data, coords):
    """Subselects xarray DataSet or DataArray object to provided coords. Raises exception if fails.

    Raises
    ------
    ValueError
        If coords name are not available in data

    KeyError
        If coords dims are not available in data

    Returns
    -------
    data: xarray
        xarray.DataSet or xarray.DataArray object, same type as input
    """
    try:
        return data.sel(**coords)

    except ValueError:
        invalid_coords = set(coords.keys()) - set(data.coords.keys())
        raise ValueError("Coords {} are invalid coordinate keys".format(invalid_coords))

    except KeyError as err:
        raise KeyError(
            (
                "Coords should follow mapping format {{coord_name:[dim1, dim2]}}. "
                "Check that coords structure is correct and"
                " dimensions are valid. {}"
            ).format(err)
        )


def color_from_dim(dataarray, dim_name):
    """Return colors and color mapping of a DataArray using coord values as color code.

    Parameters
    ----------
    dataarray : xarray.DataArray
    dim_name : str
        dimension whose coordinates will be used as color code.

    Returns
    -------
    colors : array of floats
        Array of colors (as floats for use with a cmap) for each element in the dataarray.
    color_mapping : mapping coord_value -> float
        Mapping from coord values to corresponding color
    """
    present_dims = dataarray.dims
    coord_values = dataarray[dim_name].values
    unique_coords = set(coord_values)
    color_mapping = {coord: num / len(unique_coords) for num, coord in enumerate(unique_coords)}
    if len(present_dims) > 1:
        multi_coords = dataarray.coords.to_index()
        coord_idx = present_dims.index(dim_name)
        colors = [color_mapping[coord[coord_idx]] for coord in multi_coords]
    else:
        colors = [color_mapping[coord] for coord in coord_values]
    return colors, color_mapping


def format_coords_as_labels(dataarray):
    """Format 1d or multi-d dataarray coords as strings."""
    coord_labels = dataarray.coords.to_index().values
    if isinstance(coord_labels[0], tuple):
        fmt = ", ".join(["{}" for _ in coord_labels[0]])
        coord_labels[:] = [fmt.format(*x) for x in coord_labels]
    else:
        coord_labels[:] = ["{}".format(s) for s in coord_labels]
    return coord_labels


def set_xticklabels(ax, coord_labels):
    """Set xticklabels to label list using Matplotlib default formatter."""
    ax.xaxis.get_major_locator().set_params(nbins=9, steps=[1, 2, 5, 10])
    xticks = ax.get_xticks().astype(np.int64)
    xticks = xticks[(xticks >= 0) & (xticks < len(coord_labels))]
    if len(xticks) > len(coord_labels):
        ax.set_xticks(np.arange(len(coord_labels)))
        ax.set_xticklabels(coord_labels)
    else:
        ax.set_xticks(xticks)
        ax.set_xticklabels(coord_labels[xticks])



from typing import Union, List, Optional

import numpy as np
import matplotlib.pyplot as plt

from gempy.core.data import GeoModel
from gempy_viewer.API._plot_2d_sections_api import plot_sections
from gempy_viewer.core.data_to_show import DataToShow
from gempy_viewer.core.section_data_2d import SectionData2D
from gempy_viewer.modules.plot_2d.multi_axis_manager import sections_iterator, orthogonal_sections_iterator
from gempy_viewer.modules.plot_2d.visualization_2d import Plot2D


def plot_2d(model: GeoModel,
            n_axis=None,
            section_names: list = None,
            cell_number: Optional[Union[int | list[int] | str | list[str]]] = None,
            direction: Optional[Union[str | list[str]]] = 'y',
            series_n: Union[int, List[int]] = 0,
            legend: bool = True,
            ve=1,
            block=None,
            override_regular_grid=None,
            kwargs_topography=None,
            kwargs_lithology=None,
            kwargs_scalar_field=None,
            **kwargs):
    """Plot 2-D sections of geomodel.

    Plot cross sections either based on custom section traces or cell number in xyz direction.
    Options to plot lithology block, scalar field or rendered surface lines.
    Input data and topography can be included.

    Args:
        show_block (bool): If True and model has been computed, plot cross section
         of the final model.
        show_values (bool): If True and model has been computed, plot cross section
         of the value... TODO need to add attribute to choose which value to be plot
        model: Geomodel object with solutions.
        n_axis (int): Subplot axis for multiple sections
        section_names (list): Names of predefined custom section traces
        cell_number (list): Position of the array to plot
        direction (str): Cartesian direction to be plotted (xyz)
        series_n (int): number of the scalar field.
        ve (float): vertical exageration
        override_regular_grid (numpy.ndarray): Numpy array of the size of model.grid.regular_grid. If provided, the regular grid will be overriden by this array.
        kwargs_topography (dict):
            * fill_contour
            * hillshade (bool): Calculate and add hillshading using elevation data
            * azdeg (float): azimuth of sun for hillshade
            * altdeg (float): altitude in degrees of sun for hillshade

    Keyword Args:
        legend (bool): If True plot legend. Default True
        show (bool): Call matplotlib show
        show_data (bool): Show original input data. Defaults to True.
        show_results (bool): If False, override show lith, show_calar, show_values
        show_lith (bool): Show lithological block volumes. Defaults to True.
        show_scalar (bool): Show scalar field isolines. Defaults to False.
        show_boundaries (bool): Show surface boundaries as lines. Defaults to True.
        show_topography (bool): Show topography on plot. Defaults to False. (Note Sep 2023) If True this is showing
         the height map instead of the geological map.

    Returns:
        :class:`gempy.plot.visualization_2d.Plot2D`: Plot2D object

    """

    if kwargs_lithology is None:
        kwargs_lithology = dict()
    if kwargs_topography is None:
        kwargs_topography = dict()
    if kwargs_scalar_field is None:
        kwargs_scalar_field = dict()

    if section_names is None and cell_number is None and direction is not None:
        cell_number = ['mid']

    show = kwargs.get('show', True)

    if block is not None:
        import warnings
        regular_grid = block
        warnings.warn('block is going to be deprecated. Use regular grid instead',
                      DeprecationWarning)

    section_names = [] if section_names is None else section_names
    section_names = np.atleast_1d(section_names)
    if cell_number is None:
        cell_number = []
    elif cell_number == 'mid':
        cell_number = ['mid']
    direction = [] if direction is None else direction

    if type(cell_number) != list:
        cell_number = [cell_number]

    if type(direction) != list:
        direction = [direction]

    if n_axis is None:
        n_axis = len(section_names) + len(cell_number)

    if type(series_n) is int:
        series_n = [series_n] * n_axis

    # * Grab from kwargs all the show arguments and create the proper class. This is for backwards compatibility
    can_show_results = model.solutions is not None
    data_to_show = DataToShow(
        n_axis=n_axis,
        show_data=kwargs.get('show_data', True),
        _show_results=kwargs.get('show_results', can_show_results),
        show_surfaces=kwargs.get('show_surfaces', True),
        show_lith=kwargs.get('show_lith', True),
        show_scalar=kwargs.get('show_scalar', False),
        show_boundaries=kwargs.get('show_boundaries', True),
        show_topography=kwargs.get('show_topography', False),
        show_section_traces=kwargs.get('show_section_traces', True),
        show_values=kwargs.get('show_values', False),
        show_block=kwargs.get('show_block', False)
    )

    # is 10 and 10 because in the ax pos is the second digit
    n_columns_ = 1 if len(section_names) + len(cell_number) < 2 else 2
    n_columns = n_columns_ * 10  # This is for the axis location syntax
    n_rows = (len(section_names) + len(cell_number)) / n_columns_

    n_columns_ = np.max([n_columns_, 1])
    n_rows = np.max([n_rows, 1])

    p = Plot2D()
    p.create_figure(cols=n_columns_, rows=n_rows, **kwargs)  # * This creates fig and axes
    section_data_list: list[SectionData2D] = sections_iterator(
        plot_2d=p,
        gempy_model=model,
        sections_names=section_names,
        n_axis=n_axis,
        n_columns=n_columns,
        ve=ve,
        projection_distance=kwargs.get('projection_distance', 0.2 * model.transform.isometric_scale)
    )

    orthogonal_section_data_list: list[SectionData2D] = orthogonal_sections_iterator(
        initial_axis=len(section_data_list),
        plot_2d=p,
        gempy_model=model,
        direction=direction,
        cell_number=cell_number,
        n_axis=n_axis,
        n_columns=n_columns,
        ve=ve,
        projection_distance=kwargs.get('projection_distance', 0.2 * model.transform.isometric_scale)
    )

    section_data_list.extend(orthogonal_section_data_list)

    plot_sections(
        gempy_model=model,
        sections_data=section_data_list,
        data_to_show=data_to_show,
        ve=ve,
        series_n=series_n,
        override_regular_grid=override_regular_grid,
        legend=legend,
        kwargs_topography=kwargs_topography,
        kwargs_scalar_field=kwargs_scalar_field,
        kwargs_lithology=kwargs_lithology
    )
    if show is True:
        p.fig.show()

    return p


def plot_section_traces(model: GeoModel, section_names: list[str] = None):
    """
    Plot section traces of section grid in 2-D topview (xy).
    """
    from gempy_viewer.modules.plot_2d.drawer_traces_2d import plot_section_traces as pst
    plot: Plot2D = plot_2d(
        model=model,
        n_axis=1,
        direction=['z'],
        cell_number=[-1],
        show_data=False,
        show_boundaries=False,
        show_lith=False,
        show=False
    )
    pst(
        gempy_model=model,
        ax=plot.axes[0],
        section_names=section_names
    )
    
    plot.fig.show()
    return pst


def plot_topology(geo_model, edges, centroids, direction="y", scale=True,
                  label_kwargs=None, edge_kwargs=None):
    """Plot the topology adjacency graph in 2-D.

        Args:
            geo_model ([type]): GemPy geomodel instance.
            edges (Set[Tuple[int, int]]): Set of topology edges.
            centroids (Dict[int, Array[int, 3]]): Dictionary of topology id's and
                their centroids.
            direction (Union["x", "y", "z", optional): Section direction.
                Defaults to "y".
            label_kwargs (dict, optional): Keyword arguments for topology labels.
                Defaults to None.
            edge_kwargs (dict, optional): Keyword arguments for topology edges.
                Defaults to None.

        """
    res = geo_model._grid.regular_grid.resolution
    if direction == "y":
        c1, c2 = (0, 2)
        e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
        e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
        d1 = geo_model._grid.regular_grid.extent[0]
        d2 = geo_model._grid.regular_grid.extent[4]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[0]
        r2 = res[2]
    elif direction == "x":
        c1, c2 = (1, 2)
        e1 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
        e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
        d1 = geo_model._grid.regular_grid.extent[2]
        d2 = geo_model._grid.regular_grid.extent[4]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[1]
        r2 = res[2]
    elif direction == "z":
        c1, c2 = (0, 1)
        e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
        e2 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
        d1 = geo_model._grid.regular_grid.extent[0]
        d2 = geo_model._grid.regular_grid.extent[2]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[0]
        r2 = res[1]

    tkw = {
        "color"              : "white",
        "fontsize"           : 13,
        "ha"                 : "center",
        "va"                 : "center",
        "weight"             : "ultralight",
        "family"             : "monospace",
        "verticalalignment"  : "center",
        "horizontalalignment": "center",
        "bbox"               : dict(boxstyle='round', facecolor='black', alpha=1),
    }
    if label_kwargs is not None:
        tkw.update(label_kwargs)

    lkw = {
        "linewidth": 1,
        "color"    : "black"
    }
    if edge_kwargs is not None:
        lkw.update(edge_kwargs)

    for a, b in edges:
        # plot edges
        x = np.array([centroids[a][c1], centroids[b][c1]])
        y = np.array([centroids[a][c2], centroids[b][c2]])
        if scale:
            x = x * e1 / r1 + d1
            y = y * e2 / r2 + d2
        plt.plot(x, y, **lkw)

    for node in np.unique(list(edges)):
        x = centroids[node][c1]
        y = centroids[node][c2]
        if scale:
            x = x * e1 / r1 + d1
            y = y * e2 / r2 + d2
        plt.text(x, y, str(node), **tkw)



from gempy_viewer.modules.plot_3d.vista import GemPyToVista
from functools import wraps


def plot_add_surface_points(func):
    @wraps(func)
    def pasp(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        surface_points, idx = func(*args, **kwargs)
        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_add_surface_points(idx)
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return surface_points
    return pasp


def plot_delete_surface_points(func):
    @wraps(func)
    def pdsp(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        surface_points = func(*args, **kwargs)
        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_delete_surface_points(args[1])
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return surface_points
    return pdsp


def plot_move_surface_points(func):
    @wraps(func)
    def pmsp(*args, **kwargs):
        # TODO: Add this thing with the indices to all the decorators of this
        # module
        if 'indices' not in kwargs:
            indices = args[1]
        else:
            indices = kwargs['indices']

        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        surface_points = func(*args, **kwargs)
        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_move_surface_points(indices)
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return surface_points
    return pmsp


def plot_add_orientation(func):
    @wraps(func)
    def pao(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        orientation, idx = func(*args, **kwargs)
        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_add_orientations(idx)
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return orientation
    return pao


def plot_delete_orientations(func):
    @wraps(func)
    def pdo(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        orientations = func(*args, **kwargs)
        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_delete_orientations(args[1])
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return orientations
    return pdo


def plot_move_orientations(func):
    @wraps(func)
    def pmo(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        orientations = func(*args, **kwargs)

        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_move_orientations(args[1])
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return orientations
    return pmo


def plot_set_topography(func):
    @wraps(func)
    def pst(*args, **kwargs):
        plot_object = kwargs.pop('plot_object') if 'plot_object' in kwargs else None
        topography = func(*args, **kwargs)

        if plot_object is not None:
            if isinstance(plot_object, GemPyToVista):
                if plot_object.live_updating is True:
                    plot_object.render_topography()
            else:
                raise AttributeError('plot_object must be one GemPy compatible plot')
        return topography
    return pst



import pytest
import numpy as np

from gempy.core.data import GeoModel, StackRelationType
import gempy as gp
from gempy.core.data.importer_helper import ImporterHelper


# TODO: Move this to gempy main repo
@pytest.fixture(scope='session')    
def one_fault_model_no_interp() -> GeoModel:
    geo_data = _one_fault_model_generator()

    return geo_data



@pytest.fixture(scope='session')
def one_fault_model_topo_solution() -> GeoModel:
    one_fault_model = _one_fault_model_generator()

    gp.set_section_grid(
        grid=one_fault_model.grid,
        section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                      'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
    )

    gp.set_topography_from_random(
        grid=one_fault_model.grid,
        fractal_dimension=1.2,
        d_z=np.array([300, 800]),
        topography_resolution=np.array([60, 60])
    )
    
    gp.compute_model(one_fault_model)
    return one_fault_model


def _one_fault_model_generator():
    data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
    path_to_data = data_path + "/data/input_data/jan_models/"
    geo_data = gp.create_geomodel(
        project_name='fault',
        extent=[0, 2000, 0, 2000, 0, 1000],
        resolution=[20, 5, 20],
        number_octree_levels=4,
        importer_helper=ImporterHelper(
            path_to_surface_points=path_to_data + "model5_surface_points.csv",
            path_to_orientations=path_to_data + "model5_orientations.csv",
            hash_surface_points="8fe9250462c3e65080818a84d29925378664f6be46301dcdb42ed4047aa3fe6f",
            hash_orientations="58d1d28be0c52dfdcedf36c9adc3b231e67d6923554159d6484dba589b0bfc5e",
        )
    )
    # %%
    # Setting and ordering the units and series:
    # 
    gp.map_stack_to_surfaces(
        geo_data,
        {
            "Fault_Series": 'fault',
            "Strat_Series": ('rock2', 'rock1')
        }
    )
    # %%
    # Define fault groups
    # TODO: Abstract this away with the old set_fault method
    geo_data.structural_frame.structural_groups[0].structural_relation = StackRelationType.FAULT
    geo_data.structural_frame.fault_relations = np.array([[0, 1], [0, 0]])
    return geo_data




from typing import Optional

import numpy as np
from matplotlib.axes import Axes
from matplotlib.ticker import FixedLocator, FixedFormatter

from gempy_viewer.core.slicer_data import SlicerData
from gempy_viewer.core.section_data_2d import SectionData2D, SectionType
from gempy.core.data import Grid, GeoModel
from gempy.core.data.grid_modules import Sections
from gempy_viewer.modules.plot_2d.visualization_2d import Plot2D
from gempy_viewer.modules.plot_2d.plot_2d_utils import make_section_xylabels, slice_cross_section
from gempy_viewer.modules.plot_2d.drawer_input_2d import _projection_params_section, _projection_params_regular_grid


def sections_iterator(plot_2d: Plot2D, gempy_model: GeoModel, sections_names: list[str],
                      n_axis: int, n_columns: int, ve: float, projection_distance: Optional[float] = None,
                      e:int =0) -> list[SectionData2D]:
    section_data_list: list[SectionData2D] = []
    for e, sec_name in enumerate(sections_names):
        # region matplotlib configuration
        # Check if a plot that fills all pixels is plotted
        _is_filled = False
        assert e < 10, 'Reached maximum of axes'

        ax_pos = (round(n_axis / 2 + 0.1)) * 100 + n_columns + e + 1
        temp_ax = create_ax_section(
            plot_2d=plot_2d,
            gempy_grid=gempy_model.grid,
            section_name=sec_name,
            ax_pos=ax_pos,
            ve=ve
        )
        # endregion 
        slicer_data: SlicerData = _projection_params_section(
            grid=gempy_model.grid,
            orientations=gempy_model.orientations.df.copy(),
            points=gempy_model.surface_points.df.copy(),
            projection_distance=projection_distance,
            section_name=sec_name
        )

        section_data_2d: SectionData2D = SectionData2D(
            section_type=SectionType.SECTION,
            slicer_data=slicer_data,
            ax=temp_ax,
            section_name=sec_name,
            cell_number=None,
            direction=None
        )

        section_data_list.append(section_data_2d)

    return section_data_list


def orthogonal_sections_iterator(initial_axis: int, plot_2d: Plot2D, gempy_model: GeoModel, direction: list[str], cell_number: list[int],
                                 n_axis: int, n_columns: int, ve: float, projection_distance: Optional[float] = None) -> list[SectionData2D]:
    section_data_list: list[SectionData2D] = []
    for e in range(len(cell_number)):
        # region matplotlib configuration
        # Check if a plot that fills all pixels is plotted
        _is_filled = False
        assert e < 10, 'Reached maximum of axes'

        ax_pos = (round(n_axis / 2 + 0.1)) * 100 + n_columns + e + 1 + initial_axis
        temp_ax = create_axes_orthogonal(
            plot_2d=plot_2d,
            gempy_grid=gempy_model.grid,
            cell_number=cell_number[e],
            direction=direction[e],
            ax_pos=ax_pos,
            ve=ve
        )
        
        # endregion 
        slicer_data: SlicerData = _projection_params_regular_grid(
            regular_grid=gempy_model.grid.regular_grid,
            orientations=gempy_model.orientations.df.copy(),
            points=gempy_model.surface_points.df.copy(),
            projection_distance=projection_distance,
            cell_number=cell_number[e],
            direction=direction[e]
        )

        section_data_2d: SectionData2D = SectionData2D(
            section_type=SectionType.ORTHOGONAL,
            slicer_data=slicer_data,
            ax=temp_ax,
            section_name=None,
            cell_number=cell_number[e],
            direction=direction[e]
        )

        section_data_list.append(section_data_2d)

    return section_data_list


def create_ax_section(plot_2d: Plot2D, gempy_grid: Grid, section_name, ax=None, ax_pos=111, ve=1.):
    if ax is None:
        ax = plot_2d.fig.add_subplot(ax_pos)

    if section_name is not None:
        if section_name == 'topography':
            _setup_topography_section(ax)
            extent_val = gempy_grid.topography.extent
        else:
            # Check if section is in the grid and if not raise an error
            if section_name not in gempy_grid.sections.df.index:
                raise ValueError('Section name not in grid')
            dist = gempy_grid.sections.df.loc[section_name, 'dist']
            _setup_section(
                ax=ax,
                sections=gempy_grid.sections,
                section_name=section_name,
                dist=dist
            )
            extent_val = [0, dist, gempy_grid.regular_grid.extent[4], gempy_grid.regular_grid.extent[5]]

    if extent_val[3] < extent_val[2]:  # correct vertical orientation of plot
        ax.invert_yaxis()
    plot_2d._aspect = (extent_val[3] - extent_val[2]) / (extent_val[1] - extent_val[0]) / ve
    ax.set_xlim(extent_val[0], extent_val[1])
    ax.set_ylim(extent_val[2], extent_val[3])

    ax.set_aspect('equal')
    ax.set_aspect(ve)
    
    # Adding some properties to the axes to make easier to plot
    ax.section_name = section_name
    ax.tick_params(axis='x', labelrotation=30)
    plot_2d.axes = np.append(plot_2d.axes, ax)
    plot_2d.fig.tight_layout()

    return ax


def create_axes_orthogonal(plot_2d: Plot2D, gempy_grid: Grid, cell_number,
                           direction, ax=None, ax_pos=111, ve=1.):
    if ax is None:
        ax = plot_2d.fig.add_subplot(ax_pos)

    extent_val = _setup_orthogonal_section(ax, cell_number, direction, gempy_grid)

    if extent_val[3] < extent_val[2]:  # correct vertical orientation of plot
        ax.invert_yaxis()
    # plot_2d._aspect = (extent_val[3] - extent_val[2]) / (extent_val[1] - extent_val[0]) / ve
    ax.set_xlim(extent_val[0], extent_val[1])
    ax.set_ylim(extent_val[2], extent_val[3])

    # ax.set_aspect('equal')

    # Adding some properties to the axes to make easier to plot
    ax.cell_number = cell_number
    ax.direction = direction
    ax.tick_params(axis='x', labelrotation=30)
    plot_2d.axes = np.append(plot_2d.axes, ax)
    plot_2d.fig.tight_layout()

    return ax


def _setup_orthogonal_section(ax, cell_number, direction, gempy_grid):
    _a, _b, _c, extent_val, x, y = slice_cross_section(gempy_grid.regular_grid, direction, cell_number)[:-2]  # * This requires the grid object
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    ax.set(title='Cell Number: ' + str(cell_number) + ' Direction: ' + str(direction))
    return extent_val


def _setup_section(ax: Axes, sections: Sections, section_name: str, dist: float):
    labels, axname = make_section_xylabels(
        sections=sections,
        section_name=section_name,
        n=len(ax.get_xticklabels()) - 2
    )
    pos_list = np.linspace(0, dist, len(labels))
    ax.xaxis.set_major_locator(FixedLocator(nbins=len(labels), locs=pos_list))
    ax.xaxis.set_major_formatter(FixedFormatter((labels)))
    ax.set(title=section_name, xlabel=axname, ylabel='Z')


def _setup_topography_section(ax):
    ax.set_title('Geological map')
    ax.set_xlabel('X')
    ax.set_ylabel('Y')



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify it under the
    terms of the GNU General Public License as published by the Free Software
    Foundation, either version 3 of the License, or (at your option) any later
    version.

    gempy is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
    details.

    You should have received a copy of the GNU General Public License along with
    gempy.  If not, see <http://www.gnu.org/licenses/>.


    Module with classes and methods to visualized structural geology data and
    potential fields of the regional modelling based on the potential field
    method. Tested on Ubuntu 18, MacOS 12

    Created on 16.10.2019

    @author: Miguel de la Varga, Bane Sullivan, Alexander Schaaf
"""

import warnings
from typing import Union, Dict, List, Iterable, Set, Tuple

import matplotlib.colors as mcolors
from matplotlib import cm
import numpy as np
import pandas as pn

from gempy_viewer.optional_dependencies import require_gempy

try:
    import pyvista as pv
    import pyvistaqt as pvqt

    PYVISTA_IMPORT = True
except ImportError:
    PYVISTA_IMPORT = False

warnings.filterwarnings(
    action="ignore",
    message='.*Conversion of the second argument of issubdtype *.',
    append=True
)

try:
    import vtk
    from vtk.util.numpy_support import numpy_to_vtk

    VTK_IMPORT = True
except ImportError:
    VTK_IMPORT = False

from logging import debug

try:
    from gempy.core.solution import Solution
    from gempy.core.data.geo_model import GeoModel
except ImportError:
    Solution = None
    GeoModel = None

# ? (miguel, Jun 2023) Cannot we depend on gempy? The dependency graph should go in that direction i.e. gempy_viewer depends on gempy


class __Vista:
    """
    Class to visualize data and results in 3D. Init will create all the render properties while the method render
    model will lunch the window. Using set_surface_points, set_orientations and set_surfaces in between can be chosen what
    will be displayed.
    """

    def __init__(self, model, extent=None, lith_c: pn.DataFrame = None, real_time=False,
                 plotter_type='basic', **kwargs):

        # Override default notebook value
        kwargs['notebook'] = kwargs.get('notebook', True)

        self.model = model
        self.extent = model._grid.regular_grid.extent if extent is None else extent
        self._color_lot = model._surfaces.df.set_index('id')['color'] if lith_c is None else lith_c

        self.s_widget = pn.DataFrame(columns=['val'])
        self.p_widget = pn.DataFrame(columns=['val'])
        self.surf_polydata = pn.DataFrame(columns=['val'])

        self.vista_rgrids_mesh = {}
        self.vista_rgrids_actors = {}
        self.vista_topo_actors = {}
        self.vista_surf_actor = {}

        self.real_time = real_time

        if plotter_type == 'basic':
            self.p = pv.Plotter(**kwargs)
        elif plotter_type == 'background':
            self.p = pvqt.BackgroundPlotter(**kwargs)

        self.set_bounds()
        self.p.view_isometric(negative=False)

    def update_colot_lot(self, lith_c=None):
        if lith_c is None:
            lith_c = self.model._surfaces.df.set_index('id')['color'] if lith_c is None else lith_c
            # Hopefully this removes the colors that exist in surfaces but not in data
            idx_uniq = self.model._surface_points.df['id'].unique()
            # + basement
            idx = np.append(idx_uniq, idx_uniq.max() + 1)
            lith_c = lith_c[idx]
        self._color_lot = lith_c

    def set_bounds(self, extent=None, grid=False, location='furthest', **kwargs):
        if extent is None:
            extent = self.extent
        self.p.show_bounds(bounds=extent, location=location, grid=grid, **kwargs)

    def plot_structured_grid(
            self,
            regular_grid=None,
            data: Union[dict, str] = 'Default',
            name='lith',
            **kwargs
    ):
        # Remove previous actor with the same name:
        try:
            self.p.remove_actor(self.vista_rgrids_actors[name])
        except KeyError:
            pass

        self.update_colot_lot()
        if regular_grid is None:
            regular_grid = self.model._grid.regular_grid

        g_values = regular_grid.values
        g_3D = g_values.reshape(*regular_grid.resolution, 3).T
        rg = pv.StructuredGrid(*g_3D)

        self.plot_scalar_data(rg, data, name)
        if name == 'lith':
            n_faults = self.model._faults.df['isFault'].sum()
            cmap = mcolors.ListedColormap(list(self._color_lot[n_faults:]))

            kwargs['cmap'] = kwargs.get('cmap', cmap)

        self.vista_rgrids_mesh[name] = rg

        actor = self.p.add_mesh(rg, **kwargs)
        self.vista_rgrids_actors[name] = actor
        return actor

    def plot_scalar_data(self, regular_grid, data: Union[dict, Solution, str] = 'Default', name='lith'):
        """

        Args:
            regular_grid:
            data: dictionary or solution
            name: if data is a gp.Solutions object, name of the grid that you want to plot.

        Returns:

        """
        if data == 'Default':
            data = self.model.solutions
        if isinstance(data, Solution):
            if name == 'lith':
                data = {'lith': data.lith_block}

            elif name == 'scalar':
                data = {name: data.scalar_field_matrix.T}

            elif name == 'values':
                data = {name: data.values_matrix.T}

        if type(data) == dict:
            for key in data:
                regular_grid.point_arrays[key] = data[key]

        return regular_grid

    def call_back_sphere(self, *args):

        new_center = args[0]
        obj = args[-1]
        # Get which sphere we are moving
        index = obj.WIDGET_INDEX
        try:
            self.call_back_sphere_change_df(index, new_center)
            self.call_back_sphere_move_changes(index)

        except KeyError as e:
            print('call_back_sphere error:', e)

        if self.real_time:
            try:
                self.update_surfaces_real_time()

            except AssertionError:
                print('Not enough data to compute the model')

    # region Surface Points
    def call_back_sphere_change_df(self, index, new_center):
        index = np.atleast_1d(index)
        # Modify Pandas DataFrame
        self.model.modify_surface_points(index, X=[new_center[0]], Y=[new_center[1]], Z=[new_center[2]])

    def call_back_sphere_move_changes(self, indices):
        df_changes = self.model._surface_points.df.loc[np.atleast_1d(indices)][['X', 'Y', 'Z', 'id']]
        for index, df_row in df_changes.iterrows():
            new_center = df_row[['X', 'Y', 'Z']].values

            # Update renderers
            s1 = self.s_widget.loc[index, 'val']
            r_f = s1.GetRadius() * 2
            s1.PlaceWidget(new_center[0] - r_f, new_center[0] + r_f,
                           new_center[1] - r_f, new_center[1] + r_f,
                           new_center[2] - r_f, new_center[2] + r_f)

            s1.GetSphereProperty().SetColor(mcolors.hex2color(self._color_lot[df_row['id']]))

    def call_back_sphere_delete_point(self, ind_i):
        """
        Warning this deletion system will lead to memory leaks until the vtk object is reseted (hopefully). This is
        mainly a vtk problem to delete objects
        """
        ind_i = np.atleast_1d(ind_i)
        # Deactivating widget
        for i in ind_i:
            self.s_widget.loc[i, 'val'].Off()

        self.s_widget.drop(ind_i)

    # endregion

    def plot_data(self, surface_points=None, orientations=None, **kwargs):
        # TODO: When we call this method we need to set the plotter.notebook to False!
        self.plot_surface_points(surface_points, **kwargs)
        self.plot_orientations(orientations, **kwargs)

    def plot_surface_points(self, surface_points=None, radius=None, clear=True, **kwargs):
        self.update_colot_lot()
        if clear is True:
            self.p.clear_sphere_widgets()

        # Calculate default surface_points radius
        if radius is None:
            _e = self.extent
            _e_dx = _e[1] - _e[0]
            _e_dy = _e[3] - _e[2]
            _e_dz = _e[5] - _e[4]
            _e_d_avrg = (_e_dx + _e_dy + _e_dz) / 3
            radius = _e_d_avrg * .01

        if surface_points is None:
            surface_points = self.model._surface_points.df

        test_callback = True if self.real_time is True else False

        # This is Bane way. It gives me some error with index slicing
        centers = surface_points[['X', 'Y', 'Z']]
        colors = self._color_lot[surface_points['id']].values
        s = self.p.add_sphere_widget(self.call_back_sphere,
                                     center=centers, color=colors, pass_widget=True,
                                     test_callback=test_callback,
                                     indices=surface_points.index.values,
                                     radius=radius, **kwargs)

        self.s_widget.append(pn.DataFrame(data=s, index=surface_points.index, columns=['val']))

        return self.s_widget

    def call_back_plane(self, normal, origin, obj):
        """
              Function that rules what happens when we move a plane. At the moment we update the other 3 renderers and
              update the pandas data frame
              """
        # Get new position of the plane and GxGyGz
        new_center = origin
        new_normal = normal

        # Get which plane we are moving
        index = obj.WIDGET_INDEX

        self.call_back_plane_change_df(index, new_center, new_normal)
        # TODO: rethink why I am calling this. Technically this happens outside. It is for sanity check?
        self.call_back_plane_move_changes(index)

        if self.real_time:

            try:
                self.update_surfaces_real_time()

            except AssertionError:
                print('Not enough data to compute the model')

        return True

    def call_back_plane_change_df(self, index, new_center, new_normal):
        # Modify Pandas DataFrame
        # update the gradient vector components and its location
        self.model.modify_orientations(index, X=new_center[0], Y=new_center[1], Z=new_center[2],
                                       G_x=new_normal[0], G_y=new_normal[1], G_z=new_normal[2])
        return True

    def call_back_plane_move_changes(self, indices):
        df_changes = self.model._orientations.df.loc[np.atleast_1d(indices)][['X', 'Y', 'Z',
                                                                              'G_x', 'G_y', 'G_z', 'id']]
        for index, new_values_df in df_changes.iterrows():
            new_center = new_values_df[['X', 'Y', 'Z']].values
            new_normal = new_values_df[['G_x', 'G_y', 'G_z']].values
            new_source = vtk.vtkPlaneSource()
            new_source.SetCenter(new_center)
            new_source.SetNormal(new_normal)
            new_source.Update()

            plane1 = self.p_widget.loc[index, 'val']
            #  plane1.SetInputData(new_source.GetOutput())
            plane1.SetNormal(new_normal)
            plane1.SetCenter(new_center[0], new_center[1], new_center[2])

            plane1.GetPlaneProperty().SetColor(
                pv.Color(self.model._surfaces.df.set_index('id')['color'][new_values_df['id']]).float_rgb)  # self.C_LOT[new_values_df['id']])
            plane1.GetHandleProperty().SetColor(
                pv.Color(self.model._surfaces.df.set_index('id')['color'][new_values_df['id']]).float_rgb)
        return True

    def plot_orientations(self, orientations=None, clear=True, **kwargs):
        self.update_colot_lot()
        if clear is True:
            self.p.clear_plane_widgets()
        factor = kwargs.get('factor', 0.1)
        kwargs['test_callback'] = kwargs.get('test_callback', False)

        if orientations is None:
            orientations = self.model._orientations.df
        for e, val in orientations.iterrows():
            c = self._color_lot[val['id']]
            p = self.p.add_plane_widget(self.call_back_plane,
                                        implicit=False, pass_widget=True,
                                        normal=val[['G_x', 'G_y', 'G_z']],
                                        origin=val[['X', 'Y', 'Z']], color=c,
                                        bounds=self.model._grid.regular_grid.extent,
                                        factor=factor, **kwargs)
            p.WIDGET_INDEX = e
            self.p_widget.at[e] = p
        return self.p_widget

    def plot_surfaces(self, surfaces=None, delete_surfaces=True, **kwargs):
        self.update_colot_lot()
        if delete_surfaces is True:
            for actor in self.vista_surf_actor.values():
                self.delete_surface(actor)

        if surfaces is None:
            surfaces = self.model._surfaces.df

        select_active = surfaces['isActive']
        for idx, val in surfaces[select_active][['vertices', 'edges', 'color']].dropna().iterrows():
            surf = pv.PolyData(val['vertices'], np.insert(val['edges'], 0, 3, axis=1).ravel())
            self.surf_polydata.at[idx] = surf
            self.vista_surf_actor[idx] = self.p.add_mesh(surf, pv.Color(val['color']).float_rgb, **kwargs)

        self.set_bounds()
        return self.surf_polydata

    def update_surfaces(self):
        surfaces = self.model._surfaces
        # TODO add the option of update specific surfaces
        for idx, val in surfaces.df[['vertices', 'edges', 'color']].dropna().iterrows():
            self.surf_polydata.loc[idx, 'val'].points = val['vertices']
            self.surf_polydata.loc[idx, 'val'].faces = np.insert(val['edges'], 0, 3, axis=1).ravel()

        return True

    def add_surface(self):
        raise NotImplementedError

    def delete_surface(self, actor):
        self.p.remove_actor(actor)
        return True

    def update_surfaces_real_time(self, delete=True):
        # TODO (@miguel, Jun 2023): we need a dependency injection here
        raise NotImplementedError

        try:
            gp.compute_model(self.model, sort_surfaces=False, compute_mesh=True)
        except IndexError:
            print('IndexError: Model not computed. Laking data in some surface')
        except AssertionError:
            print('AssertionError: Model not computed. Laking data in some surface')

        self.update_surfaces()
        return True

    def plot_topography(self, topography=None, scalars='geo_map', **kwargs):
        """

        Args:
            topography: gp Topography object, np.array or None
            scalars:
            **kwargs:

        Returns:

        """
        if topography is None:
            topography = self.model._grid.topography.values
        rgb = False

        # Create vtk object
        cloud = pv.PolyData(topography)

        # Set scalar values
        if scalars == 'geo_map':
            arr_ = np.empty((0, 3), dtype='int')

            # Convert hex colors to rgb
            for val in list(self._color_lot):
                rgb = (255 * np.array(mcolors.hex2color(val)))
                arr_ = np.vstack((arr_, rgb))

            sel = np.round(self.model.solutions.geological_map[0]).astype(int)[0]

            scalars_val = numpy_to_vtk(arr_[sel - 1], array_type=3)
            cm = None
            rgb = True

        elif scalars == 'topography':
            scalars_val = topography[:, 2]
            cm = 'terrain'

        elif type(scalars) is np.ndarray:
            scalars_val = scalars
            scalars = 'custom'
            cm = 'terrain'
        else:
            raise AttributeError()

        topo_actor = self.p.add_mesh(cloud.delaunay_2d(), scalars=scalars_val, cmap=cm, rgb=rgb, **kwargs)
        self.vista_topo_actors[scalars] = topo_actor
        return topo_actor


class Vista:
    def __init__(
            self,
            model: GeoModel,
            extent: List[float] = None,
            color_lot: pn.DataFrame = None,
            real_time: bool = False,
            lith_c: pn.DataFrame = None,
            plotter_type: Union["basic", "background"] = 'basic',
            **kwargs
    ):
        """GemPy 3-D visualization using pyVista.
        
        Args:
            model (gp.Model): Geomodel instance with solutions.
            extent (List[float], optional): Custom extent. Defaults to None.
            color_lot (pn.DataFrame, optional): Custom color scheme in the form
                of a look-up table. Defaults to None.
            real_time (bool, optional): Toggles real-time updating of the plot. 
                Defaults to False.
            plotter_type (Union["basic", "background"], optional): Set the 
                plotter type. Defaults to 'basic'.
        """
        self.model = model
        if extent:
            self.extent = list(extent)
        else:
            self.extent = list(model._grid.regular_grid.extent)

        if plotter_type == 'basic':
            self.p = pv.Plotter(**kwargs)
        elif plotter_type == 'background':
            self.p = pvqt.BackgroundPlotter(**kwargs)

        self._surface_actors = {}
        self._surface_points_actors = {}
        self._orientations_actors = {}

        self._actors = []
        self._live_updating = False

        self.topo_edges = None
        self.topo_ctrs = None

        self.set_bounds()
        self.p.view_isometric(negative=False)

    def show(self):
        self.p.show()

    def _actor_exists(self, new_actor):
        if not hasattr(new_actor, "points"):
            return False
        for actor in self._actors:
            actor_str = actor.points.tostring()
            if new_actor.points.tostring() == actor_str:
                debug("Actor already exists.")
                return True
        return False

    def _get_color_lot(self, lith_c=None, faults=True):
        if lith_c is None:
            surf_df = self.model._surfaces.df.set_index('surface')
            unique_surf_points = np.unique(self.model._surface_points.df['id'])

            if len(unique_surf_points) != 0:
                bool_surf_points = np.zeros(surf_df.shape[0], dtype=bool)
                bool_surf_points[unique_surf_points - 1] = True

                surf_df['isActive'] = (surf_df['isActive'] | bool_surf_points)

                if faults is True:

                    lith_c = surf_df.groupby('isActive').get_group(True)['color']
                else:
                    lith_c = surf_df.groupby(('isActive', 'isFault')).get_group((True, False))['color']

            #
            # # Hopefully this removes the colors that exist in surfaces but not in data

            #
            # # + basement
            # idx = np.append(idx_uniq, idx_uniq.max()+1)
            # lith_c = lith_c[idx]
        print(lith_c)
        self._color_lot = lith_c
        return self._color_lot

    def _update_color_lot(self, lith_c=None):
        if lith_c is None:
            active_surfaces = self.model._surfaces.df.groupby('isActive').get_group(True)
            lith_c = active_surfaces.set_index('surface')['color'] if lith_c is None else lith_c
            #
            # # Hopefully this removes the colors that exist in surfaces but not in data
            # idx_uniq = self.model.surface_points.df['surface'].unique()
            #
            # # + basement
            # idx = np.append(idx_uniq, idx_uniq.max()+1)
            # lith_c = lith_c[idx]
        print(lith_c)
        self._color_lot = lith_c
        return self._color_lot

    def set_bounds(
            self,
            extent: list = None,
            grid: bool = False,
            location: str = 'furthest',
            **kwargs
    ):
        """Set and toggle display of bounds of geomodel.
        
        Args:
            extent (list, optional): [description]. Defaults to None.
            grid (bool, optional): [description]. Defaults to False.
            location (str, optional): [description]. Defaults to 'furthest'.
        """
        if extent is None:
            extent = self.extent
        self.p.show_bounds(
            bounds=extent, location=location, grid=grid, **kwargs
        )

    def plot_surface_points(self, fmt: str = None, render_points_as_spheres=True,
                            colors=None, **kwargs):
        if fmt is None:
            return self._plot_surface_points_all()

        i = self.model._surface_points.df.groupby("surface").groups[fmt]
        if len(i) == 0:
            return False

        mesh = pv.PolyData(
            self.model._surface_points.df.loc[i][["X", "Y", "Z"]].values
        )
        if self._actor_exists(mesh):
            return []
        if colors is None:
            colors = self._get_color_lot()

        self.p.add_mesh(
            mesh,
            color=colors[fmt],
            render_points_as_spheres=render_points_as_spheres,
            reset_camera=False,
            **kwargs
        )
        self._actors.append(mesh)
        self.set_bounds()
        return [mesh]

    def plot_orientations(self, fmt: str = None, length: float = None,
                          colors=None, **kwargs):
        if fmt is None:
            return self._plot_orientations_all()

        meshes = []
        i = self.model._orientations.df.groupby("surface").groups[fmt]
        if len(i) == 0:
            return meshes
        if not length:
            length = abs(
                np.min(
                    [
                        np.diff(self.extent[:2]),
                        np.diff(self.extent[2:4]),
                        np.diff(self.extent[4:])
                    ]
                )
            ) / 10

        pts = self.model._orientations.df.loc[i][["X", "Y", "Z"]].values
        nrms = self.model._orientations.df.loc[i][["G_x", "G_y", "G_z"]].values
        if colors is None:
            colors = self._get_color_lot()

        line_kwargs = dict(
            color=colors[fmt],
            line_width=3,
        )
        line_kwargs.update(kwargs)

        for pt, nrm in zip(pts, nrms):
            mesh = pv.Line(
                pointa=pt,
                pointb=pt + length * nrm,
            )
            if self._actor_exists(mesh):
                continue
            self.p.add_mesh(
                mesh,
                **line_kwargs
            )
            self._actors.append(mesh)
            meshes.append(mesh)
        return meshes

    def _plot_surface_points_all(self, **kwargs):
        colors = self._get_color_lot()
        meshes = []
        # for fmt in self.model.surfaces.df.surface:
        #     if fmt.lower() == "basement":
        #         continue
        for fmt in self.model._surfaces.df.groupby('hasData').get_group(True)['surface']:
            new_meshes = self.plot_surface_points(fmt, colors=colors, **kwargs)
            for mesh in new_meshes:
                if mesh is not None:
                    meshes.append(mesh)
        return meshes

    def _plot_orientations_all(self, **kwargs):
        colors = self._get_color_lot()
        meshes = []
        for fmt in self.model._surfaces.df.surface:
            if fmt.lower() == "basement":
                continue
            orient_meshes = self.plot_orientations(fmt, colors=colors, **kwargs)
            for orient_mesh in orient_meshes:
                if orient_mesh is not None:
                    meshes.append(orient_mesh)
        return meshes

    def get_surface(self, fmt: str) -> pv.PolyData:
        i = self.model._surfaces.df.index[np.where(self.model._surfaces.df.surface == fmt)[0][0]]
        ver = self.model.solutions.vertices[i]

        sim = self._simplices_to_pv_tri_simplices(
            self.model.solutions.edges[i]
        )
        mesh = pv.PolyData(ver, sim)
        return mesh

    def plot_surface(self, fmt: str = None, colors=None, **kwargs):
        if colors is None:
            colors = self._get_color_lot()
        mesh = self.get_surface(fmt)
        if self._actor_exists(mesh):
            return []

        mesh_kwargs = dict(color=colors[fmt])
        mesh_kwargs.update(kwargs)

        self.p.add_mesh(mesh, **mesh_kwargs)
        self._actors.append(mesh)
        self._surface_actors[fmt] = mesh
        return [mesh]

    def clip_horizon_with_faults(
            self,
            horizon: pv.PolyData,
            faults: Iterable[pv.PolyData],
            value: float = None
    ) -> List[pv.PolyData]:
        """Clip given horizon surface with given list of fault surfaces. The
        given value represents the distance to clip away from the fault 
        surfaces.
        
        Args:
            horizon (pv.PolyData): The horizon surface to be clipped.
            faults (Iterable[pv.PolyData]): Fault(s) surface(s) to clip with.
            value (float, optional): Set the clipping value of the implicit 
                function (clipping distance from faults). Defaults to 50.
        
        Returns:
            List[pv.PolyData]: Individual clipped horizon surfaces.
        """
        if hasattr(faults, "next"):
            if type(faults[0]) == str:
                faults = [self.get_surface(f) for f in faults]

        horizons = []
        if not value:
            value = np.mean(self.model._grid.regular_grid.get_dx_dy_dz()[:2])

        # TODO: this somehow doesn't work properly with Gullfaks model
        horizons.append(
            horizon.clip_surface(faults[0], value=-value)
        )

        horizons.append(
            horizon.clip_surface(faults[-1], invert=False, value=-value)
        )

        if len(faults) == 1:
            print("Returning after 1")
            return horizons

        for f1, f2 in zip(faults[:-1], faults[1:]):
            horizons.append(
                horizon.clip_surface(
                    f1, invert=False, value=value
                ).clip_surface(
                    f2, value=-value
                )
            )

        return horizons

    def plot_surfaces(self, fmts: Iterable[str] = None, **kwargs):
        """Plot all geomodel surfaces. If given an iterable containing surface
        strings, it will plot all surfaces specified in it.
        
        Args:
            fmts (List[str], optional): Names of surfaces to plot. 
                Defaults to None.
        """
        colors = self._get_color_lot()
        meshes = []
        if fmts is None:
            fmts = self.model._surfaces.df.surface[:-1].values
        fmts = np.atleast_1d(fmts)
        for fmt in fmts:
            m = self.plot_surface(fmt, colors=colors, **kwargs)
            for mesh in m:
                meshes.append(mesh)
        return meshes

    @staticmethod
    def _simplices_to_pv_tri_simplices(sim: np.ndarray) -> np.ndarray:
        """Convert triangle simplices (n, 3) to pyvista-compatible
        simplices (n, 4).

        Args:
            sim (numpy.ndarray[int, ..., 3])

        Returns:
              numpy.ndarray[int, ..., 4]

        """
        n_edges = np.ones(sim.shape[0]) * 3
        return np.append(n_edges[:, None], sim, axis=1)

    def plot_structured_grid(self, name: str, series: str = None, render_topography: bool = False,
                             **kwargs) -> list:
        """Plot a structured grid of the geomodel.

        Args:
            name (str): Can be either one of the following

                'lith' - Lithology id block.
                'scalar' - Scalar field block.
                'values' - Values matrix block.
        """
        regular_grid = self.model._grid.regular_grid

        grid_values = regular_grid.values
        grid_3d = grid_values.reshape(*regular_grid.resolution, 3).T
        mesh = pv.StructuredGrid(*grid_3d)

        if name == "lith":
            vals = self.model.solutions.lith_block.copy()
            n_faults = self.model._faults.df['isFault'].sum()
            cmap = mcolors.ListedColormap(list(self._get_color_lot(faults=False)))
            kwargs['cmap'] = kwargs.get('cmap', cmap)
        elif name == "scalar":
            if series == None:
                # default to oldest series above basement
                series = self.model._stack.df.iloc[-2].name
            vals = self.model.solutions.scalar_field_matrix.copy()[
                self.model._stack.df.index.get_loc(series)]
        elif name == "values":
            vals = self.model.solutions.values_matrix.copy().T
            if vals.shape[1] == 0:
                print("No scalar values matrix found in given geomodel.")
                return

        mesh.point_arrays[name] = vals

        if render_topography == True:
            mesh[name][regular_grid.mask_topo.T.ravel(order='F')] = -100
            mesh = mesh.threshold(-99)

        if self._actor_exists(mesh):
            return []
        self._actors.append(mesh)
        self.p.add_mesh(mesh, **kwargs)
        return [mesh]

    def set_scalar_data(self, regular_grid, data: Union[dict, Solution, str] = 'Default', name='lith'):
        """

        Args:
            regular_grid:
            data: dictionary or solution
            name: if data is a gp.Solutions object, name of the grid that you want to plot.

        Returns:

        """
        if data == 'Default':
            data = self.model.solutions

        if isinstance(data, Solution):
            if name == 'lith':
                data = {'lith': data.lith_block}

            elif name == 'scalar':
                data = {name: data.scalar_field_matrix.T}

            elif name == 'values':
                data = {name: data.values_matrix.T}

        if type(data) == dict:
            for key in data:
                regular_grid.point_arrays[key] = data[key]

        return regular_grid

    def plot_structured_grid_interactive(
            self,
            name: str,
            render_topography: bool = False,
            **kwargs,
    ):
        """Plot interactive 3-D geomodel with three cross-sections in subplot.

        Args:
            geo_model: Geomodel object with solutions.
            name (str): Can be either one of the following
                'lith' - Lithology id block.
                'scalar' - Scalar field block.
                'values' - Values matrix block.
            render_topography: Render topography. Defaults to False.
            **kwargs:

        Returns:
            (Vista) GemPy Vista object for plotting.
        """
        mesh = self.plot_structured_grid(name=name, render_topography=render_topography, **kwargs)[0]

        # define colormaps
        if name == "lith":
            cmap = mcolors.ListedColormap(list(self._get_color_lot(faults=False)))
        elif name == "scalar":
            cmap = cm.viridis

        # callback functions for subplots
        def xcallback(normal, origin):
            self.p.subplot(1)
            self.p.add_mesh(mesh.slice(normal=normal, origin=origin), name="xslc", cmap=cmap)

        def ycallback(normal, origin):
            self.p.subplot(2)
            self.p.add_mesh(mesh.slice(normal=normal, origin=origin), name="yslc", cmap=cmap)

        def zcallback(normal, origin):
            self.p.subplot(3)
            self.p.add_mesh(mesh.slice(normal=normal, origin=origin), name="zslc", cmap=cmap)

        # cross section widgets
        self.p.subplot(0)
        self.p.add_plane_widget(xcallback, normal="x")
        self.p.subplot(0)
        self.p.add_plane_widget(ycallback, normal="y")
        self.p.subplot(0)
        self.p.add_plane_widget(zcallback, normal="z")

        # Lock other three views in place
        self.p.subplot(1)
        self.p.view_yz()
        self.p.disable()
        self.p.subplot(2)
        self.p.view_xz()
        self.p.disable()
        self.p.subplot(3)
        self.p.view_xy()
        self.p.disable()

    def _callback_surface_points(self, pos, index, widget):
        i = index
        x, y, z = pos

        self.model.modify_surface_points(i, X=x, Y=y, Z=z)

        if self._live_updating:
            self._recompute()
            self._update_surface_polydata()

    def _callback_orientations(self, normal, loc, widget):
        i = widget.WIDGET_INDEX
        x, y, z = loc
        gx, gy, gz = normal

        self.model.modify_orientations(
            i,
            X=x, Y=y, Z=z,
            G_x=gx, G_y=gy, G_z=gz
        )

        if self._live_updating:
            self._recompute()
            self._update_surface_polydata()

    def _recompute(self, **kwargs):
        # TODO(miguel jun 2023: here we need a dependency injection
        raise NotImplementedError
        gp.compute_model(self.model, compute_mesh=True, **kwargs)
        
    def _update_surface_polydata(self):
        surfaces = self.model._surfaces.df
        for surf, (idx, val) in zip(
                surfaces.surface,
                surfaces[['vertices', 'edges']].dropna().iterrows()
        ):
            polydata = self._surface_actors.get(surf, False)
            if polydata:
                polydata.points = val["vertices"]
                polydata.faces = np.insert(
                    val['edges'], 0, 3, axis=1
                ).ravel()
                self._surface_actors[surf] = polydata

    def plot_surface_points_interactive(self, fmt: str, **kwargs):
        self._live_updating = True
        i = self.model._surface_points.df.groupby("surface").groups[fmt]
        if len(i) == 0:
            return

        pts = self.model._surface_points.df.loc[i][["X", "Y", "Z"]].values

        self.p.add_sphere_widget(
            self._callback_surface_points,
            center=pts,
            radius=np.mean(self.extent) / 20,
            color=self._get_color_lot()[fmt],
            indices=i,
            test_callback=False,
            phi_resolution=6,
            theta_resolution=6,
            pass_widget=True,
            **kwargs
        )

    def plot_surface_points_interactive_all(self, **kwargs):
        self._live_updating = True
        for fmt in self.model._surfaces.df.surface:
            if fmt.lower() == "basement":
                continue
            self.plot_surface_points_interactive(fmt, **kwargs)

    def plot_orientations_interactive(self, fmt: str):
        self._live_updating = True
        i = self.model._orientations.df.groupby("surface").groups[fmt]
        if len(i) == 0:
            return

        pts = self.model._orientations.df.loc[i][["X", "Y", "Z"]].values
        nrms = self.model._orientations.df.loc[i][["G_x", "G_y", "G_z"]].values

        for index, pt, nrm in zip(i, pts, nrms):
            widget = self.p.add_plane_widget(
                self._callback_orientations,
                normal=nrm,
                origin=pt,
                bounds=self.extent,
                factor=0.15,
                implicit=False,
                pass_widget=True,
                test_callback=False,
                color=self._get_color_lot()[fmt]
            )
            widget.WIDGET_INDEX = index

    def plot_orientations_interactive_all(self):
        self._live_updating = True
        for fmt in self.model._surfaces.df.surface:
            if fmt.lower() == "basement":
                continue
            self.plot_orientations_interactive(fmt)

    def _scale_topology_centroids(
            self,
            centroids: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        """Scale topology centroid coordinates from grid coordinates to 
        physical coordinates.
        
        Args:
            centroids (Dict[int, Array[float, 3]]): Centroid dictionary.
        
        Returns:
            Dict[int, Array[float, 3]]: Rescaled centroid dictionary.
        """
        res = self.model._grid.regular_grid.resolution
        scaling = np.diff(self.extent)[::2] / res

        scaled_centroids = {}
        for n, pos in centroids.items():
            pos_scaled = pos * scaling
            pos_scaled[0] += np.min(self.extent[:2])
            pos_scaled[1] += np.min(self.extent[2:4])
            pos_scaled[2] += np.min(self.extent[4:])
            scaled_centroids[n] = pos_scaled

        return scaled_centroids

    def plot_topology(
            self,
            edges: Set[Tuple[int, int]],
            centroids: Dict[int, np.ndarray],
            node_kwargs: dict = {},
            edge_kwargs: dict = {}
    ):
        """Plot geomodel topology graph based on given set of topology edges 
        and node centroids.
        
        Args:
            edges (Set[Tuple[int, int]]): Topology edges.
            centroids (Dict[int, Array[float, 3]]): Topology node centroids
            node_kwargs (dict, optional): Node plotting options. Defaults to {}.
            edge_kwargs (dict, optional): Edge plotting options. Defaults to {}.
        """
        
        gp_plugins = require_gempy()
        lot = gp_plugins.assets.topology.get_lot_node_to_lith_id(self.model, centroids)
        centroids_scaled = self._scale_topology_centroids(centroids)
        colors = self._get_color_lot()
        for node, pos in centroids_scaled.items():
            mesh = pv.Sphere(
                center=pos,
                radius=np.average(self.extent) / 15
            )
            # * Requires topo id to lith id lot
            self.p.add_mesh(
                mesh,
                color=colors.iloc[lot[node] - 1],
                **node_kwargs
            )

        ekwargs = dict(
            line_width=3
        )
        ekwargs.update(edge_kwargs)

        for e1, e2 in edges:
            pos1, pos2 = centroids_scaled[e1], centroids_scaled[e2]

            x1, y1, z1 = pos1
            x2, y2, z2 = pos2
            x, y, z = (x1, x2), (y1, y2), (z1, z2)
            pos_mid = (
                min(x) + (max(x) - min(x)) / 2,
                min(y) + (max(y) - min(y)) / 2,
                min(z) + (max(z) - min(z)) / 2
            )
            mesh = pv.Line(
                pointa=pos1,
                pointb=pos_mid,
            )
            self.p.add_mesh(mesh, color=colors.iloc[lot[e1] - 1], **ekwargs)

            mesh = pv.Line(
                pointa=pos_mid,
                pointb=pos2,
            )
            self.p.add_mesh(mesh, color=colors.iloc[lot[e2] - 1], **ekwargs)

    def plot_topography(
            self,
            topography=None,
            scalars="geomap",
            **kwargs
    ):
        if not topography:
            try:
                topography = self.model._grid.topography.values
            except AttributeError:
                print("Unable to plot topography: Given geomodel instance "
                      "does not contain topography grid.")
                return

        polydata = pv.PolyData(topography)

        rgb = False
        if scalars == "geomap":
            arr_ = np.empty((0, 3), dtype=int)
            # convert hex colors to rgb
            for val in list(self._get_color_lot(faults=False)):
                rgb = (255 * np.array(mcolors.hex2color(val)))
                arr_ = np.vstack((arr_, rgb))

            sel = np.round(self.model.solutions.geological_map[0]).astype(int)[0]

            scalars_val = numpy_to_vtk(arr_[sel - 1], array_type=3)
            cm = None
            rgb = True
        elif scalars == "topography":
            scalars_val = topography[:, 2]
            cm = 'terrain'
        elif type(scalars) is np.ndarray:
            scalars_val = scalars
            scalars = 'custom'
            cm = 'terrain'
        else:
            raise AttributeError("Parameter scalars needs to be either \
                'geomap', 'topography' or a np.ndarray with scalar values")

        topography_actor = self.p.add_mesh(
            polydata.delaunay_2d(),
            scalars=scalars_val,
            cmap=cm,
            rgb=rgb,
            **kwargs
        )
        self._surface_actors["topography"] = topography_actor
        return topography_actor

    def plot_scalar_surfaces_3D(self, surfaces_nr: int = 10):
        """Plot scalar field as surfaces

        Args:
            surfaces_nr: Number of plotted scalar field isosurfaces

        Returns:

        """
        regular_grid = self.model._grid.regular_grid

        grid_values = regular_grid.values
        grid_3d = grid_values.reshape(*regular_grid.resolution, 3).T
        mesh = pv.StructuredGrid(*grid_3d)

        values = self.model.solutions.scalar_field_matrix.reshape(self.model._grid.regular_grid.resolution)
        mesh["vol"] = values.flatten()
        contours = mesh.contour(np.linspace(values.min(), values.max(), surfaces_nr + 2))
        self.p.add_mesh(contours, show_scalar_bar=True, label="scalar_field_main")



import enum


class ScalarDataType(enum.Enum):
    LITHOLOGY = enum.auto()
    SCALAR_FIELD = enum.auto()
    VALUES = enum.auto()
    ALL = enum.auto()
    

class TopographyDataType(enum.Enum):
    TOPOGRAPHY = enum.auto()
    GEOMAP = enum.auto()
    SCALARS = enum.auto()  # Not implemented yet


import warnings
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from gempy.core.data import GeoModel
from gempy_viewer.core.data_to_show import DataToShow
from gempy_viewer.core.section_data_2d import SectionData2D, SectionType
from ..modules.plot_2d.drawer_contours_2d import plot_regular_grid_contacts
from ..modules.plot_2d.drawer_input_2d import draw_data
from ..modules.plot_2d.drawer_regular_grid_2d import plot_section_area, plot_regular_grid_area
from ..modules.plot_2d.drawer_scalar_field_2d import plot_section_scalar_field, plot_regular_grid_scalar_field
from ..modules.plot_2d.drawer_topography_2d import plot_topography
from ..modules.plot_2d.drawer_traces_2d import plot_section_traces
from ..modules.plot_2d.plot_2d_utils import get_geo_model_cmap, get_geo_model_norm


def plot_sections(gempy_model: GeoModel, sections_data: list[SectionData2D], data_to_show: DataToShow,
                  ve: float = 1, series_n: Optional[list[int]] = None, override_regular_grid: Optional[np.ndarray] = None,
                  legend: bool = True,
                  kwargs_topography: dict = None,
                  kwargs_scalar_field: dict = None,
                  kwargs_lithology: dict = None
                  ):
    kwargs_lithology = kwargs_lithology if kwargs_lithology is not None else {}
    kwargs_scalar_field = kwargs_scalar_field if kwargs_scalar_field is not None else {}
    kwargs_topography = kwargs_topography if kwargs_topography is not None else {}
    
    series_n = series_n if series_n is not None else [0]
    
    legend_already_added = False

    for e, section_data in enumerate(sections_data):
        temp_ax = section_data.ax
        # region plot methods
        if data_to_show.show_data[e] is True:
            draw_data(
                ax=temp_ax,
                surface_points_colors=gempy_model.structural_frame.surface_points_colors_per_item,
                orientations_colors=gempy_model.structural_frame.orientations_colors_per_item,
                orientations=gempy_model.orientations.df.copy(),
                points=gempy_model.surface_points.df.copy(),
                slicer_data=section_data.slicer_data
            )

        if data_to_show.show_lith[e] is True:
            _is_filled = True
            cmap = kwargs_lithology.get('cmap', get_geo_model_cmap(gempy_model.structural_frame.elements_colors))
            norm = kwargs_lithology.get('norm', get_geo_model_norm(gempy_model.structural_frame.number_of_elements))
            
            match section_data.section_type:
                case SectionType.SECTION:
                    plot_section_area(
                        gempy_model=gempy_model,
                        ax=temp_ax,
                        section_name=section_data.section_name,
                        cmap=cmap,
                        norm=norm,
                    )
                case SectionType.ORTHOGONAL:
                    if override_regular_grid is None:
                        block_to_plot = gempy_model.solutions.raw_arrays.lith_block
                        cmap = cmap
                        norm = norm
                    else:
                        block_to_plot = override_regular_grid
                        cmap = kwargs_lithology.get('cmap', None)
                        norm = kwargs_lithology.get('norm', None)

                    plot_regular_grid_area(
                        ax=temp_ax,
                        slicer_data=section_data.slicer_data,
                        block=block_to_plot,  # * Only used for orthogonal sections
                        resolution=gempy_model.grid.regular_grid.resolution,
                        cmap=cmap,
                        norm=norm,
                    )
                case _:
                    raise ValueError(f'Unknown section type: {section_data.section_type}')
        if data_to_show.show_scalar[e] is True:
            _is_filled = True
            match section_data.section_type:
                case SectionType.SECTION:
                    plot_section_scalar_field(
                        gempy_model=gempy_model,
                        ax=temp_ax,
                        section_name=section_data.section_name,
                        series_n=series_n[e],
                        kwargs=kwargs_scalar_field
                    )
                case SectionType.ORTHOGONAL:
                    plot_regular_grid_scalar_field(
                        ax=temp_ax,
                        slicer_data=section_data.slicer_data,
                        block=gempy_model.solutions.raw_arrays.scalar_field_matrix[series_n[e]],
                        resolution=gempy_model.grid.regular_grid.resolution,
                        kwargs=kwargs_scalar_field
                    )
                case _:
                    raise ValueError(f'Unknown section type: {section_data.section_type}')
        if data_to_show.show_boundaries[e] is True:
            match section_data.section_type:
                case SectionType.SECTION:
                    warnings.warn(
                        message='Section contacts not implemented yet. We need to pass scalar field for the sections grid',
                        category=UserWarning
                    )
                    pass
                    # plot_section_contacts()
                case SectionType.ORTHOGONAL:
                    plot_regular_grid_contacts(
                        gempy_model=gempy_model,
                        ax=temp_ax,
                        slicer_data=section_data.slicer_data,
                        resolution=gempy_model.grid.regular_grid.resolution,
                        only_faults=False,
                        kwargs=kwargs_topography
                    )
                case _:
                    raise ValueError(f'Unknown section type: {section_data.section_type}')

        if data_to_show.show_topography[e] is True:
            plot_topography(
                gempy_model=gempy_model,
                ax=temp_ax,
                fill_contour=kwargs_topography.get('fill_contour', True),
                section_name=section_data.section_name,
                **kwargs_topography
            )

            if data_to_show.show_section_traces is True and section_data.section_name == 'topography':
                plot_section_traces(
                    gempy_model=gempy_model,
                    ax=temp_ax,
                    section_names=[section_data.section_name for section_data in sections_data],
                )

        # TODO: Revive the other solutions
        # elif data_to_show.show_values[e] is True: # and model.solutions.values_matrix.shape[0] != 0:
        #     _is_filled = True
        #     p.plot_values(temp_ax, series_n=series_n[e], section_name=sn, **kwargs)
        # elif show_block[e] is True and model.solutions.block_matrix.shape[0] != 0:
        #     _is_filled = True
        #     p.plot_block(temp_ax, series_n=series_n[e], section_name=sn, **kwargs)
        
        # endregion

        if legend and not legend_already_added:
            colors = gempy_model.structural_frame.elements_colors_contacts

            markers = [plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in colors]
            temp_ax.legend(
                markers,
                gempy_model.structural_frame.elements_names,
                numpoints=1
            )
            legend_already_added = True

            try:
                temp_ax.legend_.set_frame_on(True)
                temp_ax.legend_.set_zorder(10000)
            except AttributeError:
                pass

        if ve != 1:
            temp_ax.set_aspect(ve)

    return



from dataclasses import dataclass
from matplotlib import colors as mcolors, cm


@dataclass
class Plotting2DOptions:
    cmap: mcolors.ListedColormap = cm.viridis
    norm: mcolors.Normalize = mcolors.Normalize(vmin=0, vmax=1)
    
    def __post_init__(self):
        self.cmap.set_under(color="white")
        self.cmap.set_over(color="black")
        
        self.norm.clip = True
        self.norm.autoscale_None = False
        

    # ? Do I need _custom_colormap field


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


Module with classes and methods to visualized structural geology data and potential fields of the regional modelling based on
the potential field method.

@author: Miguel de la Varga
"""

import warnings

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns

sns.set_context('talk')
plt.style.use(['seaborn-v0_8-white', 'seaborn-v0_8-talk'])

warnings.filterwarnings("ignore", message="No contour levels were found")


# ! Missing a function? Check gempy_legacy
class Plot2D:
    # _color_lot: dict
    axes: list[plt.Axes]

    def __init__(self):
        # TODO: Moving this to plotting options
        self.axes = list()

    @staticmethod
    def remove(ax):
        while len(ax.collections) != 0:
            list(map(lambda x: x.remove(), ax.collections))

    def create_figure(self, figsize=None, textsize=None, **kwargs):
        """
        Create the figure.

        Args:
            figsize:
            textsize:

        Returns:
            figure, list axes, subgrid values
        """
        cols = kwargs.get('cols', 1)
        rows = kwargs.get('rows', 1)

        figsize, self.ax_labelsize, _, self.xt_labelsize, self.linewidth, _ = _scale_fig_size(
            figsize,
            textsize,
            rows,
            cols
        )
        self.fig = plt.figure(figsize=figsize, constrained_layout=False)
        self.fig.is_legend = False

        return self.fig, self.axes  # , self.gs_0


def _scale_fig_size(figsize, textsize, rows=1, cols=1):
    """Scale figure properties according to rows and cols.

    Parameters
    ----------
    figsize : float or None
        Size of figure in inches
    textsize : float or None
        fontsize
    rows : int
        Number of rows
    cols : int
        Number of columns

    Returns
    -------
    figsize : float or None
        Size of figure in inches
    ax_labelsize : int
        fontsize for axes label
    titlesize : int
        fontsize for title
    xt_labelsize : int
        fontsize for axes ticks
    linewidth : int
        linewidth
    markersize : int
        markersize
    """
    params = matplotlib.rcParams
    rc_width, rc_height = tuple(params["figure.figsize"])
    rc_ax_labelsize = params["axes.labelsize"]
    rc_titlesize = params["axes.titlesize"]
    rc_xt_labelsize = params["xtick.labelsize"]
    rc_linewidth = params["lines.linewidth"]
    rc_markersize = params["lines.markersize"]
    if isinstance(rc_ax_labelsize, str):
        rc_ax_labelsize = 15
    if isinstance(rc_titlesize, str):
        rc_titlesize = 16
    if isinstance(rc_xt_labelsize, str):
        rc_xt_labelsize = 14

    if figsize is None:
        width, height = rc_width, rc_height
        sff = 1 if (rows == cols == 1) else 1.2
        width = width * cols * sff
        height = height * rows * sff
    else:
        width, height = figsize

    if textsize is not None:
        scale_factor = textsize / rc_xt_labelsize
    elif rows == cols == 1:
        scale_factor = ((width * height) / (rc_width * rc_height)) ** 0.5
    else:
        scale_factor = 1

    ax_labelsize = rc_ax_labelsize * scale_factor
    titlesize = rc_titlesize * scale_factor
    xt_labelsize = rc_xt_labelsize * scale_factor
    linewidth = rc_linewidth * scale_factor
    markersize = rc_markersize * scale_factor

    return (width, height), ax_labelsize, titlesize, xt_labelsize, linewidth, markersize



from typing import Union

import numpy as np
import pandas as pd

from gempy_viewer.modules.plot_3d.vista import GemPyToVista


# ? Is this used?
def select_surfaces_data(data_df: pd.DataFrame, surfaces: Union[str, list[str]] = 'all') -> pd.DataFrame:
    """Select the surfaces that has to be plot.

    Args:
        data_df (pd.core.frame.DataFrame): GemPy data df that contains
            surface property. E.g Surfaces, SurfacePoints or Orientations.
        surfaces: If 'all' select all the active data. If a list of surface
            names or a surface name is passed, plot only those.
    """
    if surfaces == 'all':
        geometric_data = data_df
    else:
        geometric_data = pd.concat([data_df.groupby('surface').get_group(group) for group in surfaces])
    return geometric_data


def set_scalar_bar(gempy_vista: GemPyToVista, elements_names: list[str], surfaces_ids: np.ndarray):
    import pyvista as pv
    
    # Get mapper actor 
    if gempy_vista.surface_points_actor is not None:
        mapper_actor: pv.Actor = gempy_vista.surface_points_actor
    elif gempy_vista.regular_grid_actor is not None:
        mapper_actor = gempy_vista.regular_grid_actor
    else:
        return None  # * Not a good mapper for the scalar bar
    
    annotations = {}
    for e, name in enumerate(elements_names):
        annotations[e] = name

    mapper_actor.mapper.lookup_table.annotations = annotations
    
    sargs = gempy_vista.scalar_bar_arguments
    sargs["mapper"] = mapper_actor.mapper
    
    gempy_vista.p.add_scalar_bar(**sargs)
    gempy_vista.p.update_scalar_bar_range((surfaces_ids.min(), surfaces_ids.max())) # * This has to be here to now screw the colors with the volumes



import matplotlib
import numpy as np

from gempy.core.data import GeoModel
from gempy_viewer.core.slicer_data import SlicerData


def plot_regular_grid_contacts(gempy_model: GeoModel, ax: matplotlib.axes.Axes, slicer_data: SlicerData, resolution: iter,
                               only_faults: bool = False, **kwargs):
    if only_faults:
        raise NotImplementedError('Only faults not implemented yet')
    #     contour_idx = list(self.model._faults.df[self.model._faults.df['isFault'] == True].index)
    # else:
    #     contour_idx = list(self.model._surfaces.df.index)

    zorder = kwargs.get('zorder', 100)

    shape = resolution
    c_id = 0  # * color id startpoint
    all_colors = gempy_model.structural_frame.elements_colors_contacts

    for e, block in enumerate(gempy_model.solutions.raw_arrays.scalar_field_matrix):
        _scalar_field_per_surface = np.where(gempy_model.solutions.raw_arrays.scalar_field_at_surface_points[e] != 0)
        level = gempy_model.solutions.raw_arrays.scalar_field_at_surface_points[e][_scalar_field_per_surface]
        c_id2 = c_id + len(level)

        color_list = all_colors[c_id:c_id2]

        image = block.reshape(shape)[
            slicer_data.regular_grid_x_idx,
            slicer_data.regular_grid_y_idx,
            slicer_data.regular_grid_z_idx
        ].T

        ax.contour(
            image,
            0,
            levels=np.sort(level),
            colors=color_list[::-1],
            linestyles='solid',
            origin='lower',
            extent=[*ax.get_xlim(), *ax.get_ylim()],
            zorder=zorder - (e + len(level))
        )

        c_id = c_id2


def plot_section_contacts(ax, extent_val, section_name, zorder):
    
    # ! TODO: Update this function to use the new solutions
    if section_name == 'topography':
        shape = self.model._grid.topography.resolution

        scalar_fields = self.model.solutions.geological_map[1:]
        c_id = 0  # color id startpoint

        for e, block in enumerate(scalar_fields):
            level = self.model.solutions.scalar_field_at_surface_points[e][np.where(
                self.model.solutions.scalar_field_at_surface_points[e] != 0)]

            c_id2 = c_id + len(level)  # color id endpoint
            ax.contour(
                block.reshape(shape).T, 0,
                levels=np.sort(level),
                colors=self.cmap.colors[c_id:c_id2][::-1],
                linestyles='solid',
                origin='lower',
                extent=extent_val,
                zorder=zorder - (e + len(level))
            )
            c_id = c_id2

    else:
        l0, l1 = self.model._grid.sections.get_section_args(section_name)
        shape = self.model._grid.sections.df.loc[section_name, 'resolution']
        scalar_fields = self.model.solutions.sections[1:][:, l0:l1]

        c_id = 0  # color id startpoint

        for e, block in enumerate(scalar_fields):
            level = self.model.solutions.scalar_field_at_surface_points[e][np.where(
                self.model.solutions.scalar_field_at_surface_points[e] != 0)]

            # Ignore warning about some scalars not being on the plot since it is very common
            # that an interface does not exit for a given section
            c_id2 = c_id + len(level)  # color id endpoint
            color_list = self.model._surfaces.df.groupby('isActive').get_group(True)['color'][c_id:c_id2][::-1]

            ax.contour(block.reshape(shape).T, 0, levels=np.sort(level),
                       # colors=self.cmap.colors[self.model.surfaces.df['isActive']][c_id:c_id2],
                       colors=color_list,
                       linestyles='solid', origin='lower',
                       extent=extent_val, zorder=zorder - (e + len(level))
                       )
            c_id = c_id2



from typing import Optional

import numpy as np
import pyvista as pv
from vtkmodules.util.numpy_support import numpy_to_vtk
import matplotlib.colors as mcolors

from gempy.core.data.grid_modules import Topography
from gempy_viewer.core.scalar_data_type import TopographyDataType
from gempy_engine.core.data.raw_arrays_solution import RawArraysSolution
from gempy_viewer.modules.plot_3d.vista import GemPyToVista


def plot_topography_3d(
        gempy_vista: GemPyToVista,
        topography: Topography,
        solution: Optional[RawArraysSolution],
        topography_scalar_type: TopographyDataType,
        elements_colors: list[str],
        contours=True,
        **kwargs
):
    rgb = False

    xx, yy = np.meshgrid(topography.y, topography.x)

    grid = pv.StructuredGrid(yy, xx, topography.values_2d[:, :, 2])
    polydata = grid.extract_surface()

    if solution is None:
        is_geological_map = False
    else:
        geological_map: np.array = solution.geological_map
        is_geological_map = not(geological_map is None or geological_map.size == 0)
    
    match topography_scalar_type, is_geological_map:
        case TopographyDataType.GEOMAP, True:
            colors_hex = elements_colors
            colors_rgb_ = [list(mcolors.hex2color(val)) for val in colors_hex]  # Convert hex to RGB using list comprehension

            colors_rgb = np.array(colors_rgb_) * 255  # Multiply by 255 to get RGB values in [0, 255]
            sel = np.round(geological_map).astype(int) - 1
            selected_colors = colors_rgb[sel]  # Use numpy advanced indexing to get the corresponding RGB values
            scalars_val = numpy_to_vtk(selected_colors, array_type=3)  # Convert to vtk array

            cm = mcolors.ListedColormap(elements_colors)
            rgb = True

            show_scalar_bar = False
            scalars = 'id'
            clim = (0, len(elements_colors) - 1),

        case TopographyDataType.SCALARS, True:
            raise NotImplementedError('Not implemented yet')
            clim = None
        case _:  # * Plot topography 
            scalars_val = topography.values[:, 2]
            cm = 'terrain'
            clim = None
            show_scalar_bar = True
            scalars = 'height'

    polydata['id'] = scalars_val
    polydata['height'] = topography.values[:, 2]

    topography_actor = gempy_vista.p.add_mesh(
        polydata,
        scalars=scalars,
        cmap=cm,
        rgb=rgb,
        show_scalar_bar=False,
        clim= clim,
        **kwargs
    )

    if contours is True:
        contours = polydata.contour(scalars='height')
        contours_actor = gempy_vista.p.add_mesh(contours, color="white", line_width=3)

        gempy_vista.surface_poly['topography'] = polydata
        gempy_vista.surface_poly['topography_cont'] = contours
        gempy_vista.surface_actors["topography"] = topography_actor
        gempy_vista.surface_actors["topography_cont"] = contours_actor
    return topography_actor



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

    Module with classes and methods to perform implicit regional modelling based on
    the potential field method.
    Tested on Ubuntu 16

    Created on 10/11/2019

    @author: Alex Schaaf, Elisa Heim, Miguel de la Varga
"""
# This is for sphenix to find the packages
# sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )

import matplotlib.pyplot as plt
import pandas as pn

try:
    import mplstereonet

    mplstereonet_import = True
except ImportError:
    mplstereonet_import = False



def plot_stereonet(self, litho=None, planes=True, poles=True,
                   single_plots=False,
                   show_density=False):
    if mplstereonet_import is False:
        raise ImportError(
            'mplstereonet package is not installed. No stereographic projection available.')

    from collections import OrderedDict

    if litho is None:
        litho = self.model._orientations.df['surface'].unique()

    if single_plots is False:
        fig, ax = mplstereonet.subplots(figsize=(5, 5))
        df_sub2 = pn.DataFrame()
        for i in litho:
            df_sub2 = df_sub2.append(self.model._orientations.df[
                                         self.model._orientations.df[
                                             'surface'] == i])

    for formation in litho:
        if single_plots:
            fig = plt.figure(figsize=(5, 5))
            ax = fig.add_subplot(111, projection='stereonet')
            ax.set_title(formation, y=1.1)

        # if series_only:
        # df_sub = self.model.orientations.df[self.model.orientations.df['series'] == formation]
        # else:
        df_sub = self.model._orientations.df[
            self.model._orientations.df['surface'] == formation]

        if poles:
            ax.pole(df_sub['azimuth'] - 90, df_sub['dip'], marker='o',
                    markersize=7,
                    markerfacecolor=self._color_lot[formation],
                    markeredgewidth=1.1, markeredgecolor='gray',
                    label=formation + ': ' + 'pole point')
        if planes:
            ax.plane(df_sub['azimuth'] - 90, df_sub['dip'],
                     color=self._color_lot[formation],
                     linewidth=1.5, label=formation + ': ' + 'azimuth/dip')
        if show_density:
            if single_plots:
                ax.density_contourf(df_sub['azimuth'] - 90, df_sub['dip'],
                                    measurement='poles', cmap='viridis',
                                    alpha=.5)
            else:
                ax.density_contourf(df_sub2['azimuth'] - 90, df_sub2['dip'],
                                    measurement='poles', cmap='viridis',
                                    alpha=.5)

        fig.subplots_adjust(top=0.8)
        handles, labels = ax.get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        ax.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1.9, 1.1))
        ax._grid(True, color='black', alpha=0.25)





import pyvista as pv
import numpy as np

from gempy.core.data.structural_element import StructuralElement
from gempy_viewer.modules.plot_3d.vista import GemPyToVista



def plot_surfaces(
        gempy_vista: GemPyToVista,
        structural_elements_with_solution: list[StructuralElement],
        **kwargs
):
    # ! If the order of the meshes does not match the order of scalar_field_at_surface points we need to reorder them in 'multi_scalar_dual_contouring.py'
    
    topography_mesh = gempy_vista.surface_poly.get('topography', None)
    
    for element in structural_elements_with_solution:
        vertices_ = element.vertices
        edges_ = element.edges
        if vertices_ is None or vertices_.shape[0] == 0 or edges_.shape[0] == 0:
            continue
        surf = pv.PolyData(vertices_, np.insert(edges_, 0, 3, axis=1).ravel())
        
        if topography_mesh is not None:
            surf = surf.clip_surface(topography_mesh, invert=True)
        
        gempy_vista.surface_poly[element.name] = surf
        gempy_vista.surface_actors[element.name] = gempy_vista.p.add_mesh(
            surf,
            pv.Color(element.color).float_rgb,
            show_scalar_bar=False,
            # cmap=cmap,
            **kwargs
        )



import numpy as np
import pyvista as pv

from gempy.core.data import GeoModel
from gempy.core.data.orientations import OrientationsTable
from gempy.core.data.surface_points import SurfacePointsTable
from gempy_viewer.modules.plot_2d.plot_2d_utils import get_geo_model_cmap
from gempy_viewer.modules.plot_3d.vista import GemPyToVista


def plot_data(gempy_vista: GemPyToVista, 
              model: GeoModel,
              arrows_factor: float,
              **kwargs):
    plot_surface_points(
        gempy_vista=gempy_vista,
        surface_points=model.surface_points,
        elements_colors=model.structural_frame.elements_colors_contacts,
        **kwargs
    )

    plot_orientations(
        gempy_vista=gempy_vista,
        orientations=model.orientations,
        elements_colors=model.structural_frame.elements_colors_orientations,
        arrows_factor=arrows_factor,
        **kwargs
    )


def plot_surface_points(
        gempy_vista: GemPyToVista,
        surface_points: SurfacePointsTable,
        elements_colors: list[str],
        render_points_as_spheres=True,
        point_size=10, **kwargs
):
    ids = surface_points.ids
    if ids.shape[0] == 0:
        return
    unique_values, first_indices = np.unique(ids, return_index=True)  # Find the unique elements and their first indices
    unique_values_order = unique_values[np.argsort(first_indices)]  # Sort the unique values by their first appearance in `a`

    mapping_dict = {value: i for i, value in enumerate(unique_values_order)}  # Use a dictionary to map the original numbers to new values
    mapped_array = np.vectorize(mapping_dict.get)(ids)  # Map the original array to the new values

    # Selecting the surfaces to plot
    xyz = surface_points.xyz
    if transfromed_data := False:  # TODO: Expose this to user
        xyz = surface_points.model_transform.apply(xyz)
        
    poly = pv.PolyData(xyz)
    poly['id'] = mapped_array

    cmap = get_geo_model_cmap(
        elements_colors=np.array(elements_colors),
        reverse=False)

    gempy_vista.surface_points_mesh = poly
    gempy_vista.surface_points_actor = gempy_vista.p.add_mesh(
        mesh=poly,
        cmap=cmap,  # TODO: Add colors
        scalars='id',
        render_points_as_spheres=render_points_as_spheres,
        point_size=point_size,
        show_scalar_bar=False
    )


def plot_orientations(
        gempy_vista: GemPyToVista,
        orientations: OrientationsTable,
        elements_colors: list[str],
        arrows_factor: float,
        **kwargs
):
    orientations_xyz = orientations.xyz
    orientations_grads = orientations.grads
    
    if orientations_xyz.shape[0] == 0:
        return
    if transfromed_data := False:
        orientations_xyz = orientations.model_transform.apply(orientations_xyz)
        orientations_grads = orientations.model_transform.transform_gradient(orientations_grads)
        arrows_factor /=  orientations.model_transform.isometric_scale
        
    poly = pv.PolyData(orientations_xyz)
    
    ids = orientations.ids
    if ids.shape[0] == 0:
        return
    unique_values, first_indices = np.unique(ids, return_index=True)  # Find the unique elements and their first indices
    unique_values_order = unique_values[np.argsort(first_indices)]  # Sort the unique values by their first appearance in `a`

    mapping_dict = {value: i for i, value in enumerate(unique_values_order)}  # Use a dictionary to map the original numbers to new values
    mapped_array = np.vectorize(mapping_dict.get)(ids)  # Map the original array to the new values
    
    poly['id'] = mapped_array
    poly['vectors'] = orientations_grads

    # TODO: I am still trying to figure out colors and ids in orientations and surface points
    cmap = get_geo_model_cmap(
        elements_colors=np.array(elements_colors),
        reverse=False
    )

    arrows = poly.glyph(
        orient='vectors',
        scale=False,
        factor=arrows_factor,
    )

    gempy_vista.orientations_actor = gempy_vista.p.add_mesh(
        mesh=arrows,
        cmap=cmap,
        show_scalar_bar=False
    )
    gempy_vista.orientations_mesh = arrows



import copy
import warnings

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
import scipy.spatial.distance as dd

from gempy_viewer.core.slicer_data import SlicerData
from gempy.core.data.grid_modules import RegularGrid, Sections
from gempy.core.data.grid_modules import Topography
from .plot_2d_utils import slice_cross_section
from .visualization_2d import Plot2D
from gempy.core.data import GeoModel
from gempy.core.data.grid import Grid


def _plot_data(plot_2d: Plot2D, gempy_model: GeoModel, ax, section_name=None, cell_number=None, direction='y',
              legend=True, projection_distance=None):
    warnings.warn('This function is deprecated. Use plot_data instead', DeprecationWarning)
    if projection_distance is None:
        # TODO: This has to be updated to the new location
        projection_distance = 0.2 * gempy_model.transform.isometric_scale

    # TODO: This are not here 
    points = gempy_model.surface_points.df.copy()
    orientations = gempy_model.orientations.df.copy()

    # TODO: This is a weird check to do this deep
    section_name, cell_number, direction = plot_2d._check_default_section(ax, section_name, cell_number, direction)

    if section_name is not None:
        slicer_data: SlicerData = _projection_params_section(
            grid=gempy_model.grid,
            orientations=orientations,
            points=points,
            projection_distance=projection_distance,
            section_name=section_name
        )
    else:
        slicer_data: SlicerData = _projection_params_regular_grid(
            regular_grid=gempy_model.grid.regular_grid,
            cell_number=cell_number,
            direction=direction,
            orientations=orientations,
            points=points,
            projection_distance=projection_distance
        )

    # ! Hack to keep the right X label. I think this has to be here before the plot
    temp_label = copy.copy(ax.xaxis.label)
    
    draw_data(
        ax=ax, 
        surface_points_colors=gempy_model.structural_frame.surface_points_colors_per_item,
        orientations_colors=gempy_model.structural_frame.orientations_colors_per_item,
        orientations=orientations,
        points=points,
        slicer_data=slicer_data
    )
    
    # region others
    if plot_2d.fig.is_legend is False and legend is True or legend == 'force':
        ax.legend(
            handles=[plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in gempy_model.structural_frame.elements_colors][::-1],
            labels=gempy_model.structural_frame.elements_names,
            numpoints=1
        )
        plot_2d.fig.is_legend = True

    ax.xaxis.label = temp_label
    try:
        ax.legend_.set_frame_on(True)
        ax.legend_.set_zorder(10000)
    except AttributeError:
        pass
    # endregion


# TODO: This could be public and the slice just a class yes!
def draw_data(ax, surface_points_colors: list[str], orientations_colors: list[str],
              orientations: pd.DataFrame, points: pd.DataFrame, slicer_data: SlicerData):
    
    _draw_surface_points(ax, points, slicer_data, surface_points_colors)
    _draw_orientations(ax, orientations, orientations_colors, slicer_data)


def _draw_orientations(ax, orientations, orientations_colors, slicer_data):
    sel_ori = orientations[slicer_data.select_projected_o]
    aspect = np.subtract(*ax.get_ylim()) / np.subtract(*ax.get_xlim())
    min_axis = 'width' if aspect < 1 else 'height'
    ax.quiver(
        sel_ori[slicer_data.x],
        sel_ori[slicer_data.y],
        sel_ori[slicer_data.Gx],
        sel_ori[slicer_data.Gy],
        pivot="tail",
        scale_units=min_axis,
        scale=30,
        color=np.array(orientations_colors)[slicer_data.select_projected_o],
        edgecolor='k',
        headwidth=8,
        linewidths=1,
        zorder=102
    )


def _draw_surface_points(ax, points, slicer_data, surface_points_colors):
    points_df = points[slicer_data.select_projected_p]
    ax.scatter(
        points_df[slicer_data.x],
        points_df[slicer_data.y],
        c=(np.array(surface_points_colors)[slicer_data.select_projected_p]),
        s=70,
        edgecolors='white',
        zorder=102
    )


def _projection_params_regular_grid(regular_grid: RegularGrid, cell_number, direction, orientations, points,
                                    projection_distance) -> SlicerData:
    # if cell_number is None or cell_number == "mid":
    #     cell_number = int(regular_grid.resolution[0] / 2)

    if direction == 'x' or direction == 'X':
        arg_ = 0
        dx = regular_grid.dx
        dir = 'X'
    elif direction == 'y' or direction == 'Y':
        arg_ = 2
        dx = regular_grid.dy
        dir = 'Y'
    elif direction == 'z' or direction == 'Z':
        arg_ = 4
        dx = regular_grid.dz
        dir = 'Z'
    else:
        raise AttributeError('Direction must be x, y, z')

    _loc = regular_grid.extent[arg_] + dx * int(regular_grid.resolution[0] / 2)
    cartesian_point_dist = points[dir] - _loc
    cartesian_ori_dist = orientations[dir] - _loc

    _a, _b, _c, _, x, y, Gx, Gy = slice_cross_section(
        regular_grid=regular_grid,
        direction=direction,
        cell_number=cell_number
    )
    select_projected_p = cartesian_point_dist < projection_distance
    select_projected_o = cartesian_ori_dist < projection_distance
    
    slice_data = SlicerData(
        x=x,
        y=y,
        Gx=Gx,
        Gy=Gy,
        select_projected_p=select_projected_p,
        select_projected_o=select_projected_o,
        regular_grid_x_idx=_a,
        regular_grid_y_idx=_b,
        regular_grid_z_idx=_c
    )

    return slice_data


def _projection_params_section(grid: Grid, orientations: 'pd.DataFrame', points: 'pd.DataFrame',
                               projection_distance: float, section_name: str) -> SlicerData:
    if section_name == 'topography':
        Gx, Gy, cartesian_ori_dist, cartesian_point_dist, x, y = _projection_params_topography(
            topography=grid.topography,
            orientations=orientations,
            points=points,
            projection_distance=projection_distance,
        )
    else:
        # Project points:
        sections: Sections = grid.sections
        shift = np.asarray(sections.df.loc[section_name, 'start'])
        end_point = np.atleast_2d(np.asarray(sections.df.loc[section_name, 'stop']) - shift)
        A_rotate = np.dot(end_point.T, end_point) / sections.df.loc[section_name, 'dist'] ** 2

        points_x_y = points[['X', 'Y']]
        orientations_x_y = orientations[['X', 'Y']]
    
        perpe_sqdist = ((np.dot(A_rotate, (points_x_y).T).T - points_x_y) ** 2).sum(axis=1)
        cartesian_point_dist = np.sqrt(perpe_sqdist)
        perpe_sqdist = ((np.dot(A_rotate, (orientations_x_y).T).T - orientations_x_y) ** 2).sum(axis=1)
        cartesian_ori_dist = np.sqrt(perpe_sqdist)

        # These are the coordinates of the data projected on the section
        cartesian_point = np.dot(A_rotate, (points_x_y - shift).T).T
        cartesian_ori = np.dot(A_rotate, (orientations_x_y - shift).T).T

        # Since we plot only the section we want the norm of those coordinates
        points['X'] = np.linalg.norm(cartesian_point, axis=1)
        orientations['X'] = np.linalg.norm(cartesian_ori, axis=1)
        x, y, Gx, Gy = 'X', 'Z', 'G_x', 'G_z'

    select_projected_p = cartesian_point_dist < projection_distance
    select_projected_o = cartesian_ori_dist < projection_distance

    slicer_data = SlicerData(
        x=x,
        y=y,
        Gx=Gx,
        Gy=Gy,
        select_projected_p=select_projected_p,
        select_projected_o=select_projected_o
    )

    return slicer_data


def _projection_params_topography(topography: Topography, orientations, points, projection_distance, topography_compression: int = 5000):
    decimation_aux = int(topography.values.shape[0] / topography_compression)
    tpp = topography.values[::decimation_aux + 1, :]
    cdist_sp = dd.cdist(
        XA=tpp,
        XB=points[['X', 'Y', 'Z']])
    cartesian_point_dist = (cdist_sp < projection_distance).sum(axis=0).astype(bool)
    cdist_ori = dd.cdist(
        XA=tpp,
        XB=orientations[['X', 'Y', 'Z']]
    )
    cartesian_ori_dist = (cdist_ori < projection_distance).sum(axis=0).astype(bool)
    x, y, Gx, Gy = 'X', 'Y', 'G_x', 'G_y'
    return Gx, Gy, cartesian_ori_dist, cartesian_point_dist, x, y



import numpy as np

from gempy_viewer.core.slicer_data import SlicerData
from gempy.core.data import GeoModel, Grid
from gempy_engine.core.data.raw_arrays_solution import RawArraysSolution
from gempy_viewer.modules.plot_2d.visualization_2d import Plot2D


def plot_regular_grid_scalar_field(ax, slicer_data: SlicerData, block: np.ndarray, resolution: iter, kwargs: dict):
    extent_val = [*ax.get_xlim(), *ax.get_ylim()]

    plot_block = block.reshape(resolution)
    image = plot_block[
        slicer_data.regular_grid_x_idx,
        slicer_data.regular_grid_y_idx,
        slicer_data.regular_grid_z_idx
    ].T


    ax.contour(
        image,
        cmap='autumn',
        extent=extent_val,
        zorder=8,
        **kwargs
    )
    if 'N' in kwargs:
        kwargs.pop('N')
    ax.contourf(
        image,
        cmap='autumn',
        extent=extent_val,
        zorder=7,
        alpha=.8,
        **kwargs
    )

    return ax


def plot_section_scalar_field(gempy_model: GeoModel, ax, section_name=None, series_n: int = 0, **kwargs):
    extent_val = [*ax.get_xlim(), *ax.get_ylim()]

    image = _prepare_section_image(gempy_model, section_name, series_n=series_n)

    ax.contour(
        image,
        cmap='autumn',
        extent=extent_val,
        zorder=8,
        **kwargs
    )
    if 'N' in kwargs:
        kwargs.pop('N')
    ax.contourf(
        image,
        cmap='autumn',
        extent=extent_val,
        zorder=7,
        alpha=.8,
        **kwargs
    )

    return ax


def _prepare_section_image(gempy_model: GeoModel, section_name: str, series_n: int = 0):
    legacy_solutions: RawArraysSolution = gempy_model.solutions.raw_arrays
    grid: Grid = gempy_model.grid

    if section_name == 'topography':
        try:
            image = legacy_solutions.geological_map[1][series_n].reshape(grid.topography.values_2d[:, :, 2].shape).T
        except AttributeError:
            raise AttributeError('Geological map not computed. Activate the topography grid.')
    else:
        assert type(section_name) == str or type(
            section_name) == np.str_, 'section name must be a string of the name of the section'
        assert legacy_solutions.sections is not None, 'no sections for plotting defined'

        l0, l1 = grid.sections.get_section_args(section_name)
        shape = grid.sections.df.loc[section_name, 'resolution']
        
        first_level_octree = gempy_model.solutions.octrees_output[0]
        group_output = first_level_octree.outputs_centers[series_n]
        scalar_field_ALL_sections = group_output.scalar_fields.exported_fields.scalar_field[group_output.grid.sections_slice]
        scalar_field_THIS_section = scalar_field_ALL_sections[l0:l1]
        image = scalar_field_THIS_section.reshape(shape[0], shape[1]).T
        # image = legacy_solutions.sections[1][l0:l1].reshape(shape[0], shape[1]).T
    return image



import sys
from .API import *

__all__ = ['plot_2d', 'plot_3d', 'plot_section_traces', 'plot_topology', 'plot_stereonet']

# Assert at least pyton 3.10
assert sys.version_info[0] >= 3 and sys.version_info[1] >= 10, "GemPy requires Python 3.10 or higher"

if __name__ == '__main__':
    pass



import matplotlib.pyplot as plt
import numpy as np
import pytest

import gempy as gp
import gempy_viewer as gpv
from gempy import GeoModel
from gempy_viewer.modules.plot_2d.visualization_2d import Plot2D
from tests.conftest import _one_fault_model_generator


# TODO: - [x] Test sections
# TODO: - [x] Refactor plotting 2d
# TODO: - [ ] Refactor public and private methods


class TestPlot2DInputData:
    @pytest.fixture(scope='class')
    def one_fault_model_no_interp(self) -> GeoModel:
        geo_data = _one_fault_model_generator()

        return geo_data

    def test_plot_2d_data_default(self, one_fault_model_no_interp):
        _: Plot2D = gpv.plot_2d(one_fault_model_no_interp)

    def test_plot_2d_data_default_all_none(self, one_fault_model_no_interp):
        gpv.plot_2d(one_fault_model_no_interp, show_data=True, show_results=False)

    def test_plot_2d_data_cross_section(self, one_fault_model_no_interp):
        geo_model = one_fault_model_no_interp
        section_dict = {'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                        'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}

        gp.set_section_grid(
            grid=geo_model.grid,
            section_dict=section_dict
        )

        gpv.plot_2d(
            model=geo_model,
            section_names=['section_SW-NE', 'section_NW-SE'],
            show_section_traces=False,  # TODO: Test this one
        )

    def test_plot_2d_topography(self, one_fault_model_no_interp):
        gp.set_topography_from_random(
            grid=one_fault_model_no_interp.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gpv.plot_2d(
            model=one_fault_model_no_interp,
            section_names=['topography'],
            show_topography=True,
            show_section_traces=False,  # TODO: Test this one
        )

    def test_plot_2d_topography_and_sections(self, one_fault_model_no_interp):
        gp.set_section_grid(
            grid=one_fault_model_no_interp.grid,
            section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                          'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
        )

        gp.set_topography_from_random(
            grid=one_fault_model_no_interp.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gpv.plot_2d(
            model=one_fault_model_no_interp,
            section_names=['section_SW-NE', 'section_NW-SE', 'topography'],
            show_topography=True,
            show_section_traces=True  # TODO: Test this one
        )

    def test_plot_2d_all_together(self, one_fault_model_no_interp):
        gp.set_section_grid(
            grid=one_fault_model_no_interp.grid,
            section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                          'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
        )

        gp.set_topography_from_random(
            grid=one_fault_model_no_interp.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gpv.plot_2d(
            model=one_fault_model_no_interp,
            section_names=['section_SW-NE', 'section_NW-SE', 'topography'],
            direction=['x'], cell_number=['mid'],
            show_topography=True,
            show_section_traces=True  # TODO: Test this one
        )


class TestPlot2DSolutions:
    @pytest.fixture(scope='class')
    def one_fault_model_topo_solution(self) -> GeoModel:
        one_fault_model = _one_fault_model_generator()

        gp.set_section_grid(
            grid=one_fault_model.grid,
            section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                          'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
        )

        gp.set_topography_from_random(
            grid=one_fault_model.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gp.compute_model(one_fault_model)
        return one_fault_model

    def test_plot_2d_solutions_default(self, one_fault_model_topo_solution):
        print(one_fault_model_topo_solution.structural_frame)
        _: Plot2D = gpv.plot_2d(one_fault_model_topo_solution)

    def test_plot_2d_all_together(self, one_fault_model_topo_solution):
        gpv.plot_2d(
            model=one_fault_model_topo_solution,
            section_names=['section_SW-NE', 'section_NW-SE', 'topography'],
            direction=['x', 'y', 'y'], cell_number=['mid', 'mid', 'mid'],
            show_lith=[False, False, False, True, True, True],
            show_boundaries=[False, False, False, True, True, True],
            show_scalar=[False, False, False, False, True, True],
            series_n=[0, 0, 0, 0, 0, 1],
            show_topography=True,
            show_section_traces=True  # TODO: Test this one
        )

    def test_ve(self, one_fault_model_topo_solution):
        raise NotImplementedError('Not implemented yet (TODO)')
        # Test ve
        p2d = gpv.plot_2d(one_fault_model_topo_solution, direction='x', cell_number='mid', show_topography=True, ve=3)

        plt.show()


class TestPlot2DSolutionsOctrees:
    @pytest.fixture(scope='class')
    def one_fault_model_topo_solution_octrees(self) -> GeoModel:
        one_fault_model = _one_fault_model_generator()
        one_fault_model.grid.regular_grid.resolution = [2, 4, 2]

        one_fault_model.interpolation_options.number_octree_levels = 5
        
        # TODO: Test octree regular grid with everything else combined
        gp.set_section_grid(
            grid=one_fault_model.grid,
            section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                          'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
        )

        gp.set_topography_from_random(
            grid=one_fault_model.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gp.compute_model(one_fault_model)
        return one_fault_model

    def test_plot_2d_solutions_default(self, one_fault_model_topo_solution_octrees):
        print(one_fault_model_topo_solution_octrees.structural_frame)
        _: Plot2D = gpv.plot_2d(
            model=one_fault_model_topo_solution_octrees,
            show_boundaries=True
        )
    
    def test_plot_2d_solutions_scalar(self, one_fault_model_topo_solution_octrees):
        print(one_fault_model_topo_solution_octrees.structural_frame)
        _: Plot2D = gpv.plot_2d(
            model=one_fault_model_topo_solution_octrees,
            show_boundaries=False,
            show_scalar=True,
            series_n=1
        )
        
        
    def test_plot_2d_all_together(self, one_fault_model_topo_solution_octrees):
        gpv.plot_2d(
            model=one_fault_model_topo_solution_octrees,
            section_names=['section_SW-NE', 'section_NW-SE', 'topography'],
            direction=['x', 'y', 'y'], cell_number=['mid', 'mid', 'mid'],
            show_lith=[False, False, False, True, True, True],
            show_boundaries=[False, False, False, True, True, True],
            show_scalar=[False, False, False, False, True, True],
            series_n=[0, 0, 0, 0, 0, 1],
            show_topography=True,
            show_section_traces=True  # TODO: Test this one
        )



from typing import Sequence

import numpy as np
import matplotlib.colors as mcolors

from gempy.core.data import Grid
from gempy.core.data.grid_modules import grid_types
from gempy.core.data.grid_modules import Sections, RegularGrid


def slice_cross_section(regular_grid: grid_types.RegularGrid, direction: str, cell_number: int or str):
    """
    Slice the 3D array (blocks or scalar field) in the specific direction selected in the plot functions

    """
    _a, _b, _c = (
        slice(0, regular_grid.resolution[0]),
        slice(0, regular_grid.resolution[1]),
        slice(0, regular_grid.resolution[2])
    )

    if direction == "x":
        cell_number = int(regular_grid.resolution[0] / 2) if cell_number == 'mid' else cell_number
        _a, x, y, Gx, Gy = cell_number, "Y", "Z", "G_y", "G_z"
        extent_val = regular_grid.extent[[2, 3, 4, 5]]
    elif direction == "y":
        cell_number = int(regular_grid.resolution[1] / 2) if cell_number == 'mid' else cell_number
        _b, x, y, Gx, Gy = cell_number, "X", "Z", "G_x", "G_z"
        extent_val = regular_grid.extent[[0, 1, 4, 5]]
    elif direction == "z":
        cell_number = int(regular_grid.resolution[2] / 2) if cell_number == 'mid' else cell_number
        _c, x, y, Gx, Gy = cell_number, "X", "Y", "G_x", "G_y"
        extent_val = regular_grid.extent[[0, 1, 2, 3]]
    else:
        raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")
    return _a, _b, _c, extent_val, x, y, Gx, Gy


def make_section_xylabels(sections: Sections, section_name, n=5):
    """
    @elisa heim
    Setting the axis labels to any combination of vertical crossections

    Args:
        section_name: name of a defined gempy crossection. See gempy.Model().grid.section
        n:

    Returns:

    """
    if n > 5:
        n = 3  # todo I don't know why but sometimes it wants to make a lot of xticks
    elif n < 0:
        n = 3

    j = np.where(sections.names == section_name)[0][0]
    startend = list(sections.section_dict.values())[j]
    p1, p2 = startend[0], startend[1]
    xy = sections.calculate_line_coordinates_2points(p1, p2, n)
    if len(np.unique(xy[:, 0])) == 1:
        labels = xy[:, 1].astype(int)
        axname = 'Y'
    elif len(np.unique(xy[:, 1])) == 1:
        labels = xy[:, 0].astype(int)
        axname = 'X'
    else:
        labels = [str(xy[:, 0].astype(int)[i]) + ',\n' + str(xy[:, 1].astype(int)[i]) for i in
                  range(xy[:, 0].shape[0])]
        axname = 'X,Y'
    return labels, axname


def check_default_section(ax, section_name, cell_number, direction):
    if section_name is None:
        try:
            section_name = ax.section_name
        except AttributeError:
            pass
    if cell_number is None:
        try:
            cell_number = ax.cell_number
            direction = ax.direction
        except AttributeError:
            pass

    return section_name, cell_number, direction


def slice_topo_4_sections(grid: Grid, p1, p2, resx, method='interp2d'):
    """
    Slices topography along a set linear section

    Args:
        :param p1: starting point (x,y) of the section
        :param p2: end point (x,y) of the section
        :param resx: resolution of the defined section
        :param method: interpolation method, 'interp2d' for cubic scipy.interpolate.interp2d
                                         'spline' for scipy.interpolate.RectBivariateSpline

    Returns:
        :return: returns x,y,z values of the topography along the section
    """
    xy = grid.sections.calculate_line_coordinates_2points(p1, p2, resx)
    z = grid.sections.interpolate_zvals_at_xy(xy, grid.topography, method)
    return xy[:, 0], xy[:, 1], z


def calculate_p1p2(regular_grid: RegularGrid, direction, cell_number):
    if direction == 'y':
        cell_number = int(regular_grid.resolution[1] / 2) if cell_number == 'mid' else cell_number

        y = regular_grid.extent[2] + regular_grid.dy * cell_number
        p1 = [regular_grid.extent[0], y]
        p2 = [regular_grid.extent[1], y]

    elif direction == 'x':
        cell_number = int(regular_grid.resolution[0] / 2) if cell_number == 'mid' else cell_number

        x = regular_grid.extent[0] + regular_grid.dx * cell_number
        p1 = [x, regular_grid.extent[2]]
        p2 = [x, regular_grid.extent[3]]

    else:
        raise NotImplementedError
    return p1, p2


def get_geo_model_cmap(elements_colors: Sequence[str], reverse: bool = True) -> mcolors.ListedColormap:
    if reverse:
        return mcolors.ListedColormap(elements_colors).reversed()
    else:
        return mcolors.ListedColormap(elements_colors)


def get_geo_model_norm(number_elements: int) -> mcolors.Normalize:
    return mcolors.Normalize(vmin=0.5, vmax=number_elements + 0.5)



def require_gempy_plugins():
    try:
        import gempy.plugins
    except ImportError:
        raise ImportError("The gempy.plugins package is required to run this function.")
    return gempy.plugins


def require_pyvista():
    try:
        import pyvista as pv
    except ImportError:
        raise ImportError("The pyvista package is required to run this function.")
    return pv


def require_skimage():
    try:
        import skimage
    except ImportError:
        raise ImportError("The skimage package is required to run this function.")
    return skimage
    


from dataclasses import dataclass
from typing import Optional

import numpy as np


@dataclass(init=True)
class SlicerData:
    x: str
    y: str
    Gx: str
    Gy: str
    select_projected_p: np.ndarray
    select_projected_o: np.ndarray
    regular_grid_x_idx: Optional[int] = None
    regular_grid_y_idx: Optional[int] = None
    regular_grid_z_idx: Optional[int] = None





import matplotlib.pyplot as plt
import numpy as np

from gempy.core.data import GeoModel
from gempy_viewer.core.data_to_show import DataToShow
from gempy_viewer.core.scalar_data_type import ScalarDataType, TopographyDataType
from gempy_viewer.modules.plot_2d.plot_2d_utils import get_geo_model_cmap
from gempy_viewer.modules.plot_3d.drawer_input_3d import plot_data
from gempy_viewer.modules.plot_3d.drawer_structured_grid_3d import plot_structured_grid
from gempy_viewer.modules.plot_3d.drawer_surfaces_3d import plot_surfaces
from gempy_viewer.modules.plot_3d.drawer_topography_3d import plot_topography_3d
from gempy_viewer.modules.plot_3d.plot_3d_utils import set_scalar_bar
from gempy_viewer.modules.plot_3d.vista import GemPyToVista

try:
    import pyvista as pv
    from gempy_viewer.modules.plot_3d._vista import Vista as Vista

    PYVISTA_IMPORT = True
except ImportError:
    PYVISTA_IMPORT = False

try:
    import mplstereonet

    mplstereonet_import = True
except ImportError:
    mplstereonet_import = False


def plot_3d(
        model: GeoModel,
        plotter_type='basic',
        active_scalar_field: str = None,
        ve=None,
        topography_scalar_type: TopographyDataType = TopographyDataType.GEOMAP,
        kwargs_pyvista_bounds=None,
        kwargs_plot_structured_grid=None,
        kwargs_plot_topography=None,
        kwargs_plot_data=None,
        kwargs_plotter=None,
        kwargs_plot_surfaces=None,
        image=False,
        show=True,
        **kwargs
) -> GemPyToVista:
    """Plot 3-D geomodel."""
    # * Grab from kwargs all the show arguments and create the proper class. This is for backwards compatibility
    can_show_results = model.solutions is not None  # and model.solutions.lith_block.shape[0] != 0
    data_to_show = DataToShow(
        n_axis=1,
        show_data=kwargs.get('show_data', True),
        _show_results=kwargs.get('show_results', can_show_results),
        show_surfaces=kwargs.get('show_surfaces', True),
        show_lith=kwargs.get('show_lith', True),
        show_scalar=kwargs.get('show_scalar', False),
        show_boundaries=kwargs.get('show_boundaries', True),
        show_topography=kwargs.get('show_topography', True),
        show_section_traces=kwargs.get('show_section_traces', True),
        show_values=kwargs.get('show_values', False),
        show_block=kwargs.get('show_block', False)
    )
    kwargs_plot_topography = kwargs_plot_topography or {}
    kwargs_plot_structured_grid = kwargs_plot_structured_grid or {}
    kwargs_plot_data = kwargs_plot_data or {}
    kwargs_plotter = kwargs_plotter or {}
    kwargs_plot_surfaces = kwargs_plot_surfaces or {}
    kwargs_pyvista_bounds = kwargs_pyvista_bounds or {}

    if image is True:
        show = True
        kwargs_plotter['off_screen'] = True
        plotter_type = 'basic'

    if model.solutions is None:
        data_to_show.show_results = False
        solutions_raw_arrays = None
    else:
        solutions_raw_arrays = model.solutions.raw_arrays

    extent: np.ndarray = model.grid.regular_grid.extent

    gempy_vista = GemPyToVista(
        extent=extent,
        plotter_type=plotter_type,
        pyvista_bounds_kwargs=kwargs_pyvista_bounds,
        **kwargs_plotter
    )

    if data_to_show.show_topography[0] is True and model.grid.topography is not None:
        plot_topography_3d(
            gempy_vista=gempy_vista,
            topography=model.grid.topography,
            solution=solutions_raw_arrays,
            topography_scalar_type=topography_scalar_type,
            elements_colors=model.structural_frame.elements_colors[::-1],
            contours=kwargs_plot_topography.get('contours', True),
            **kwargs_plot_topography
        )
        
    if data_to_show.show_boundaries[0] is True and len(solutions_raw_arrays.vertices) != 0:
        plot_surfaces(
            gempy_vista=gempy_vista,
            structural_elements_with_solution=model.structural_frame.structural_elements,
            **kwargs_plot_surfaces
        )

    if data_to_show.show_data[0] is True:
        arrow_size = kwargs_plot_data.get('arrow_size', 10)
        min_axes = np.min(np.diff(extent)[[0, 2, 4]])

        plot_data(
            gempy_vista=gempy_vista,
            model=model,
            arrows_factor=arrow_size / (100 / min_axes),
            **kwargs_plot_data
        )

    if data_to_show.show_lith[0] is True:
        plot_structured_grid(
            gempy_vista=gempy_vista,
            regular_grid=model.grid.regular_grid,
            scalar_data_type=ScalarDataType.LITHOLOGY,
            active_scalar_field="lith",
            solution=solutions_raw_arrays,
            cmap=get_geo_model_cmap(model.structural_frame.elements_colors_volumes),
            **kwargs_plot_structured_grid
        )

    if data_to_show.show_scalar[0] is True:
        plot_structured_grid(
            gempy_vista=gempy_vista,
            regular_grid=model.grid.regular_grid,
            scalar_data_type=ScalarDataType.SCALAR_FIELD,
            active_scalar_field=active_scalar_field,
            solution=solutions_raw_arrays,
            cmap='viridis',
            **kwargs_plot_structured_grid
        )

    if True: 
        set_scalar_bar(
            gempy_vista=gempy_vista,
            elements_names = model.structural_frame.elements_names,
            surfaces_ids=model.structural_frame.elements_ids - 1
        )

    if ve is not None:
        gempy_vista.p.set_scale(zscale=ve)

    fig_path: str = kwargs.get('fig_path', None)
    if fig_path is not None:
        gempy_vista.p.show(screenshot=fig_path)

    if image is True:
        gempy_vista.p.show(screenshot=True)
        img = gempy_vista.p.last_image
        plt.imshow(img)
        plt.axis('off')
        plt.show(block=False)
        gempy_vista.p.close()
        show= False

    if show is True:
        gempy_vista.p.show()

    return gempy_vista



import dataclasses
from typing import Union


@dataclasses.dataclass
class DataToShow:
    n_axis: int
    show_data: Union[bool, list] = True
    _show_results: Union[bool, list] = True
    show_surfaces: Union[bool, list] = True
    show_lith: Union[bool, list] = True
    show_scalar: Union[bool, list] = False
    show_boundaries: Union[bool, list] = True
    show_topography: Union[bool, list] = False
    show_section_traces: Union[bool, list] = True
    show_values: Union[bool, list] = False
    show_block: Union[bool, list] = False
    
    def __post_init__(self):
        if self.show_results is False:
            self.show_lith = [False]
            self.show_values = [False]
            self.show_block = [False]
            self.show_scalar = [False]
            self.show_boundaries = [False]


        if type(self.show_data) is bool:
            self.show_data = [self.show_data] * self.n_axis
        if type(self.show_lith) is bool:
            self.show_lith = [self.show_lith] * self.n_axis
        if type(self.show_values) is bool:
            self.show_values = [self.show_values] * self.n_axis
        if type(self.show_block) is bool:
            self.show_block = [self.show_block] * self.n_axis
        if type(self.show_scalar) is bool:
            self.show_scalar = [self.show_scalar] * self.n_axis
        if type(self.show_boundaries) is bool:
            self.show_boundaries = [self.show_boundaries] * self.n_axis
        if type(self.show_topography) is bool:
            self.show_topography = [self.show_topography] * self.n_axis

    @property
    def show_results(self):
        return self._show_results
    
    @show_results.setter
    def show_results(self, value):
        self._show_results = value
        if value is False:
            self.show_lith = [False]
            self.show_values = [False]
            self.show_block = [False]
            self.show_scalar = [False]
            self.show_boundaries = [False]
            


import numpy as np
import pytest

import gempy_viewer as gpv
import gempy as gp
from gempy.core.data import GeoModel
from gempy_viewer.core.scalar_data_type import TopographyDataType
from tests.conftest import _one_fault_model_generator


class TestPlot3dInputData:
    def test_plot_3d_input_data(self, one_fault_model_no_interp):
        gpv.plot_3d(one_fault_model_no_interp)


class TestPlot3DSolutions:
    def test_plot_3d_solutions_default(self, one_fault_model_topo_solution):
        gpv.plot_3d(one_fault_model_topo_solution)
    
    def test_plot_3d_solutions(self, one_fault_model_topo_solution):
        gpv.plot_3d(
            model=one_fault_model_topo_solution,
            show_scalar=False,
            show_lith=True,
            show_data=True,
            show_boundaries=True
        )
    
    def test_plot_3d_scalar_field(self, one_fault_model_topo_solution):
        gpv.plot_3d(
            model=one_fault_model_topo_solution,
            active_scalar_field="sf_1",
            show_scalar=True,
            show_lith=False
        )
    
    def test_plot_3d_solutions_topography(self, one_fault_model_topo_solution):
        gpv.plot_3d(
            model=one_fault_model_topo_solution,
            show_topography=True,
            topography_scalar_type=TopographyDataType.TOPOGRAPHY
        )
    
    def test_plot_3d_solutions_topography_geological_map(self, one_fault_model_topo_solution):
        gpv.plot_3d(
            model=one_fault_model_topo_solution,
            show_lith=True,
            show_topography=True,
            topography_scalar_type=TopographyDataType.GEOMAP
        )


class TestPlot2DSolutionsOctrees:
    @pytest.fixture(scope='class')
    def one_fault_model_topo_solution_octrees(self) -> GeoModel:
        one_fault_model = _one_fault_model_generator()
        one_fault_model.grid.regular_grid.resolution = [2, 4, 2]

        one_fault_model.interpolation_options.number_octree_levels = 5

        # TODO: Test octree regular grid with everything else combined
        gp.set_section_grid(
            grid=one_fault_model.grid,
            section_dict={'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                          'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
        )

        gp.set_topography_from_random(
            grid=one_fault_model.grid,
            fractal_dimension=1.2,
            d_z=np.array([600, 2000]),
            topography_resolution=np.array([60, 60])
        )

        gp.compute_model(one_fault_model)
        return one_fault_model


    def test_plot_3d_solutions_default(self, one_fault_model_topo_solution_octrees):
        gpv.plot_3d(one_fault_model_topo_solution_octrees)



import numpy as np

from gempy.core.data.geo_model import GeoModel
from gempy.core.data.grid_modules import Sections


def plot_section_traces(gempy_model: GeoModel, ax, section_names: list[str] = None):
    sections: Sections = gempy_model.grid.sections
    
    if section_names is None:
        section_names = list(sections.names)

    for section_name in section_names:
        if section_name not in sections.df.index:
            continue
            
        x1, y1 = np.asarray(sections.df.loc[section_name, 'start'])
        x2, y2 = np.asarray(sections.df.loc[section_name, 'stop'])
        ax.plot([x1, x2], [y1, y2], label=section_name, linestyle='--')
        ax.legend(frameon=True)



from dataclasses import dataclass
from enum import auto, Enum
from typing import Optional

from matplotlib.axes import Axes

from gempy_viewer.core.slicer_data import SlicerData


class SectionType(Enum):
    SECTION: int = auto()
    ORTHOGONAL: int = auto()
    TOPOGRAPHY: int = auto()
    
    
@dataclass
class SectionData2D:
    section_type: SectionType
    slicer_data: SlicerData
    ax: Axes
    section_name: Optional[str]
    cell_number: Optional[int]
    direction: Optional[str]


# setup.py for gempy_viewer. Requierements are numpy and matplotlib

from setuptools import setup, find_packages

version = "2023.1.0b1"


def read_requirements(file_name):
    with open(file_name, "r", encoding="utf-8") as f:
        return [line.strip() for line in f.readlines()]


setup(
    name='gempy_viewer',
    version=version,
    packages=find_packages(),
    url='',
    license='EUPL',
    author='Miguel de la Varga', 
    author_email="miguel@terranigma-solutions.com",
    description='Viewer for the geological modeling package GemPy',
    install_requires=read_requirements("requirements.txt"),
    extras_require={
        "opt": read_requirements("optional_requirements.txt"),
        "dev": read_requirements("dev-requirements.txt"),
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: GIS',
        'Programming Language :: Python :: 3.10'
    ],
    python_requires='>=3.10'
)
